/** @file token.h @brief Class Token */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TOKEN_H

#include "definitions.h"
#include "helper.h"
#include <string>
#include <iostream>
#include <fstream>
#include <assert.h>

namespace faudes {

/**
 * Tokens model atomic data for stream IO.
 *
 * A Token models a string or numeric datum that can be read from a
 * or written to a C++ stream. The class itself implements the representation of 
 * the data including its type. For section handling and actual file processing
 * see TokenReader and TokenWriter.
 *
 * @param mType
 *   faudes::TokenType of the Token
 * @param mStringValue
 *   Token value as C++-type std::string
 * @param mIntegerValue
 *   Token value as C++-type long integer
 * @param mFloatValue
 *   Token value of C++-type double
 */

class Token {
public:

  /**
   * Empty constructor, constructs None token
   */
  Token(void);
		
  /**
   * Copy constructor
   */
  Token(const Token& rToken);


  /**
   * Token destructor
   */
  ~Token(void);

  /**
   * Token types:
   */
  enum TokenType {
    None,      ///<  Invalid/empty token    
    Begin,     ///<  \<label\>   (begin of section)
    End,       ///<  \<\\label\> (end of section)
    String,    ///<  "name"      (the name may not contain a quote (") )
    Option,    ///<  +xyZ+       (the option string may not contain a "+")
    Integer,   ///<  1234        (non-negative)
    Integer16, ///<  0x12fff     ("0x" makes a number Integer16) 
    Float      ///<  -12.34      ("-" or "." make a number a float)
  };


  /**
   * Initialize None token
   * 
   */
  void SetNone(void);

  /**
   * Initialize as String token
   * 
   * @param rName
   *   String to fill the Token
   */
  void SetString(const std::string& rName);

  /**
   * Initialize as Begin token
   * 
   * @param rName
   *   Title of section to fill the Token
   */
  void SetBegin(const std::string& rName);

  /**
   * Initialize as End token
   * 
   * @param rName
   *   Title of section to fill the Token
   */
  void SetEnd(const std::string& rName);

  /**
   * Initialize as Option token
   * 
   * @param rName
   *   Option to fill the Token
   */
  void SetOption(const std::string& rName);

  /**
   * Initialize as Integer token
   *
   * @param number
   *   Number to fill the Token
   */
  void SetInteger(const long int number);

  /**
   * Initialize as Integer16 token
   *
   * @param number
   *   Number to fill the Token
   */
  void SetInteger16(const long int number);

  /**
   * Initialize as Float token
   *
   * @param number
   *   Number to fill the Token
   */
  void SetFloat(const double number);
    
  /**
   * Get integer value of a numeric token
   *
   * @return
   *   Token's integer value
   */
  long int IntegerValue(void) const {
    return(mIntegerValue);
  }

  /**
   * Get float value of a numeric token
   *
   * @return
   *   Token float value
   */
  double FloatValue(void) const {
    return(mFloatValue);
  }

  /**
   * Get string value of a name token
   *
   * @return
   *   Token's name value
   */
  const std::string& StringValue(void) const {
    return(mStringValue);
  }

  /**
   * Get token Type
   *
   * @return
   *   Token's TokenType
   */
  TokenType Type(void) const {
    return(mType);
  }

  /**
   * Read Token from input stream
   * 
   * @param pStream
   *   Pointer to std::ifstream
   * @exception Exception
   *   - ios exceptions (eg file io error)
   */
  int Read(std::istream* pStream);

  /**
   * Write Token to output stream
   *
   * @param pStream 
   *   Pointer to ostream
   * @exception Exception
   *   - ios exceptions (eg file io error,)
   */
  void Write(std::ostream* pStream); 

private:
  /** Token type */
  TokenType mType; 
		
  /** Token std::string value (if Token is of type Name, Begein or End) */
  std::string mStringValue;

  /** Token integer value (if Token is of type Integer or Integer16) */
  long int mIntegerValue;

  /** Token float value (if Token is of type Float) */
  double mFloatValue;


  /**
   * Read a number from an input file stream
   * 
   * @param pStream
   *   Reference to std::istream
   * 
   * @return 
   *   Success
   */
  bool ReadNumber(std::istream* pStream); 

  /** Read a std::string value from an input file stream
   *
   * @param pStream
   *   Reference to std::istream
   * @param stop
   *   Stop character
   * 
   * @return 
   *   Success
   */
  bool ReadString(std::istream* pStream, char stop);

  /**
   * Read (ignore) spaces and comments in an input file stream
   *
   * @param pStream
   *   Reference to std::istream
   *
   * @return
   *   Number of lines read
   */   
  int  ReadSpace(std::istream* pStream);
};

} // namespace faudes

#define TOKEN_H
#endif


