/** @file op_observercomputation.cpp

Methods to compute natural projections that exhibit the obsrver property.
The observer algorithm is elaborated in
K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event
Systems," Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
In addition, methods to compute natural projections that exhibit
output control consistency (OCC) and local control consistency (LCC) are provided. See for example
K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
*/

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include "op_observercomputation.h"
  
namespace faudes{

// calcAbstAlphObs(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents)
void calcAbstAlphObs(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents) {
	OP_DF("calcAbstAlphObs(" << rGenObs.Name() << "," << "rHighAlph, rNewHighAlph, rMapRelabeledEvents)");
	// The controllable events are separated from the cGenerator. All functions that are successively 
	// called, are defined for vGenerators
	EventSet cntrevs = rGenObs.ControllableEvents();
	calcAbstAlphObs(rGenObs, cntrevs , rHighAlph, rNewHighAlph, rMapRelabeledEvents);
	// the controllable events that have been changed by the called function are set in the cGenerator cGenObs
	rGenObs.SetControllable(cntrevs);
}

// calcAbstAlphObs(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, rMapRelabeledEvents)
void calcAbstAlphObs(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents) {
	OP_DF("calcAbstAlphObs(" << rGenObs.Name() << "," << "rControllableEvents, rHighAlph, rNewHighAlph, rMapRelabeledEvents)");
	// The function called next returns a relabeled generator and a map of relabeled transitions
	map<Transition,Idx>  changedtrans;
	calcAbstAlphObs(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, changedtrans);
	// for later use, the relabeled transitions are converted into relabeled events 
	// note that this function is accumulative, i.e., rMapRelabeledEvents need not be empty
	map<Transition,Idx>::iterator rtEndIt = changedtrans.end();
	map<Transition,Idx>::iterator rtIt =  changedtrans.begin();
	for(; rtIt != rtEndIt; rtIt++){
		if(rMapRelabeledEvents.find(rtIt->first.Ev) == rMapRelabeledEvents.end() ){// if the event does not exist, yet, a nex element in the map is created
			rMapRelabeledEvents[rtIt->first.Ev] = set<Idx>();
			if(rGenObs.Alphabet().Exists(rtIt->second) )// FIXME: there seem to be relabeled transitions that do not exist 
				rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
		} 
		else {  // a new label is inserted into the map
			if(rGenObs.Alphabet().Exists(rtIt->second) )// FIXME: there seem to be relabeled transitions that do not exist    
				rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
		}
	}
}

// calcAbstAlphObs(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTrans)
void calcAbstAlphObs(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans)
{	
	OP_DF("calcAbstAlphObs(" << rGenObs.Name() << ", rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTRans)");
	// Initialization of variables
	rNewHighAlph = rHighAlph;
	rMapChangedTrans.clear();
	Generator genDyn(rGenObs);
	map<Transition,Transition> mapChangedTransReverse;
	vector<Idx> newPartitions;
	map<Idx,Idx> mapStateToPartition;
	map<Idx, EventSet> mapRelabeledEvents;
	bool done=false;
	#ifdef DF_PLOT
		Idx iterationCount=1;
		string name;
	#endif
	// observer algorithm: In each step, a dynamic system is computed that fulfills the one-step observer condition. 
	// Iterative application of the bisimulation algorithm and the relabeling procedure yields an automaton rGenObs
	// that, together with the new high-level alphabet rNewHighAlph fulfills the observer condition. 
	while(done==false)
	{
		// compute dynamic system for Lm-observer on rGenObs
		calculateDynamicSystemObs(rGenObs, rNewHighAlph, genDyn);
		#ifdef DF_PLOT
		name = ("./Automata/Plots/" + rGenObs.Name() + "DynamicSystem_" + ToString(iterationCount));
		genDyn.DotWrite(name);
		#endif
	
		Generator genPart;
		mapStateToPartition.clear();
		newPartitions.clear();
	
		// compute coarsest quasi-congruence on the dynamic system
		calcBisimulation(genDyn, mapStateToPartition, genPart, newPartitions);
		#ifdef DF_PLOT
		name = ("./Automata/Plots/" + rGenObs.Name() + "Bisimulation_" + ToString(iterationCount));
		genPart.DotWrite(name);
		++iterationCount;
		#endif
	
		// check if quotient automaton is deterministic and free of unobservable events
		// and relabel transitions in rGenObs if necessary. The high-level alphabet is modified accordingly
		done=relabel(rGenObs, rControllableEvents, rNewHighAlph, newPartitions, mapStateToPartition, mapChangedTransReverse,  rMapChangedTrans, mapRelabeledEvents);
	}
}

// calculateDynamicSystemObs(rGen, rHighAlph, rGenDyn)
void calculateDynamicSystemObs(const vGenerator& rGen, EventSet& rHighAlph, vGenerator& rGenDyn)
{
	OP_DF("calculateDynamicSystemObs(" << rGen.Name() << "," << rHighAlph.Name() << "," << rGenDyn.Name() << ")");
	// transition relation sorted in reverse order for backwards reachability
	TransSetX2EvX1 tset_X2EvX1;
	rGen.TransRel(tset_X2EvX1);
	
	// prepare generator rGenDyn
	rGenDyn.ClearTransRel();
	rGenDyn.InjectAlphabet(rHighAlph); // all high-level events are contained in the alphabet of rGenDyn
	
	// helpers
	EventSet::Iterator eIt = rGenDyn.AlphabetBegin();
	EventSet::Iterator eItEnd = rGenDyn.AlphabetEnd();
	
	TransSetX2EvX1::Iterator tItByX2Ev;
	TransSetX2EvX1::Iterator tItByX2EvEnd;
	StateSet reach;
	StateSet::Iterator sIt;
	StateSet::Iterator sItEnd;
	TransSet::Iterator tIt;
	TransSet::Iterator tItEnd;
	map<Idx,StateSet> entryStateToLocalReach;
	set<Idx>::iterator esIt, esItEnd;
	// map that maps a state (first Idx) to a reachable entry state (second Idx) via a high-level event (third Idx)
	map<Idx,vector<pair<Idx,Idx> > > stateToEntryState;
	map<Idx,vector<pair<Idx,Idx> > >::iterator stesIt;
	map<Idx,vector<pair<Idx,Idx> > >::iterator stesItEnd;
	// label for transitions to marked states in the dynamic system
	std::string eventname = ( rGenDyn.EventSymbolTablep())->UniqueSymbol("mLabel_1");
	Idx mLabel = (rGenDyn.EventSymbolTablep())->InsEntry(eventname);
	
	// algorithm for computing the dynamic system
	
	// loop over all states of original generator
	StateSet::Iterator stateSetIt = rGen.StatesBegin();
	StateSet::Iterator stateSetItEnd = rGen.StatesEnd();
	for( ; stateSetIt != stateSetItEnd; ++stateSetIt) {
		OP_DF("calculateDynamicSystemObs: loop over all states; current state: " << rGen.StateName(*stateSetIt) 
		<< " [" << *stateSetIt << "]");
		
		// compute locally reachable states for current state
		reach.Clear();
		LocalAccessibleReach(rGen, rHighAlph, *stateSetIt, reach);
		OP_DF("calculateDynamicSystemObs: states in local reach: \n " << reach.ToString() );
		
		// check if current state (*stateSetIt) is an entry-state (a backward transition with
		// a high-level event exists. If yes, the locally reachable states (reach) are stored in
		// the entryStateToLocalReach map
		tItByX2Ev=tset_X2EvX1.BeginByX2(*stateSetIt); 
		tItByX2EvEnd=tset_X2EvX1.EndByX2(*stateSetIt); 
		for(; tItByX2Ev != tItByX2EvEnd; ++tItByX2Ev)
		{
			OP_DF("calculateDynamicSystemObs: checking transition : " << rGen.TStr(*tItByX2Ev));
			if(rHighAlph.Exists(tItByX2Ev->Ev)){
			    OP_DF("calculateDynamicSystemObs: current state is an entry-state");
    			// map entry-state to its locally reachable states
    			entryStateToLocalReach[*stateSetIt]=reach;
    			break;
			}
		}
		vector<pair<Idx,Idx> > emptyVector;
		stateToEntryState[*stateSetIt]=emptyVector;
		// loop over all states in the local reach of current state to determine which marked states
		// are locally reachable and to determine which high-level events can occur after local strings
		sIt=reach.Begin();
		sItEnd=reach.End();
		for( ; sIt != sItEnd; ++sIt) 
		{
			// check if state sIt is marked; if yes, create an m-transition (mLabel) between current state and sIt
			if(rGen.ExistsMarkedState(*sIt))
			{
			    OP_DF("calculateDynamicSystemObs: marked state " << rGen.SStr(*sIt) 
    			<<  " is locally reachable from current state");
    			rGenDyn.InsEvent(mLabel); // ??
    			rGenDyn.InsState(*stateSetIt); // ??
    			rGenDyn.InsState(*sIt); //??
    			rGenDyn.SetTransition(*stateSetIt, mLabel, *sIt);
			}
			// loop over all transitions of current state sIt to determine if high-level events are possible
			tIt    = rGen.TransRelBegin(*sIt);
			tItEnd = rGen.TransRelEnd(*sIt);
			for(; tIt != tItEnd; ++tIt)
			{
			    OP_DF("calculateDynamicSystemObs: Loop over all states in reach; checking transition: " 
    				<< rGen.TStr(*tIt));
    			if(rHighAlph.Exists(tIt->Ev))
    			{
    				OP_DF("calculateDynamicSystemObs: state " << rGen.SStr(tIt->X2) << 
    				" is an entry state and can be reached from current state" << rGen.SStr(*stateSetIt) << " by event " << rGen.EventName(tIt->Ev));
    				pair<Idx,Idx> newPair;
    				newPair.first=tIt->X2;
    				newPair.second=tIt->Ev;
    				// store the reachable entry state and the corresponding high-level event
    				stateToEntryState[*stateSetIt].push_back(newPair);
    			}
			}
		}
	}
	
	// create the transition structure of the dynamic system
	stesIt    = stateToEntryState.begin();
	stesItEnd = stateToEntryState.end();
	vector<pair <Idx,Idx> >* pPairs;
	vector<pair <Idx,Idx> >::iterator vIt;
	vector<pair <Idx,Idx> >::iterator vItEnd;
	// To construct the dynamic system, each local state has to be connected to all states in
	// the local accessible reach of entry states that can be reached via a high-level event.
	// This information is contained in the stateToEntryState map combined with the entryStateToLocalReach map.
	// iteration over all entries (X1) of map stateToEntryState
	for(; stesIt != stesItEnd; ++stesIt)
	{
    	pPairs=&(stesIt->second);
    	vIt    = (*pPairs).begin();
    	vItEnd = (*pPairs).end();
    	// loop over all pairs (Ev,X2) of current entry
    	for( ; vIt != vItEnd; ++vIt)
    	{
    		// check if transition already exists
    		if(!((rGenDyn.TransRel()).Exists(stesIt->first,vIt->second,vIt->first))) 
    		{
        		// find local reach of entry state X2 
        		StateSet* pLocalReach = &((entryStateToLocalReach.find(vIt->first))->second);
        	
        		// Add a Ev-transition from X1 to every state in the local reach of X2     
        		StateSet::Iterator lrsIt    = pLocalReach->Begin();
        		StateSet::Iterator lrsItEnd = pLocalReach->End();
        		for(; lrsIt != lrsItEnd; ++lrsIt)
        		{
        			rGenDyn.InsEvent(vIt->second); // ??
        			rGenDyn.InsState(stesIt->first); // ??
        			rGenDyn.InsState(*lrsIt); //??
        			rGenDyn.SetTransition(stesIt->first,vIt->second,*lrsIt);
        			OP_DF("calculateDynamicSystemObs: Transition added to resulting generator: " << 
        				rGenDyn.TStr(Transition(stesIt->first,vIt->second,*lrsIt))); 
        		}
    		}
    	}
	}
	OP_DF("calculateDynamicSystemObs: leaving function");
}

// calcAbstAlphObsOCC(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents)
void calcAbstAlphObsOCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents){
	OP_DF("calcAbstAlphObsOCC(" << rGenObs.Name() << "," << "rHighAlph, rNewHighAlph, rMapRelabeledEvents)");
	// The controllable events are separated from the cGenerator. All functions that are successively 
	// called, are defined for vGenerators
    map<Transition,Idx>  changedtrans;
    EventSet rControllableEvents = rGenObs.ControllableEvents();
    // The function called next returns a relabeled generator and a map of relabeled transitions
    calcAbstAlphObsOCC(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, changedtrans);
    // for later use, the relabeled transitions are converted into relabeled events 
	// note that this function is accumulative, i.e., rMapRelabeledEvents need not be empty
    map<Transition,Idx>::iterator rtEndIt = changedtrans.end();
    map<Transition,Idx>::iterator rtIt =  changedtrans.begin();
    for(; rtIt != rtEndIt; rtIt++){
        if(rMapRelabeledEvents.find(rtIt->first.Ev) == rMapRelabeledEvents.end() ){
           rMapRelabeledEvents[rtIt->first.Ev] = set<Idx>();
           if(rGenObs.Alphabet().Exists(rtIt->second) )// FIXME: there seem to be relabeled transitions that do not exist 
               rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
        } else {  
           if(rGenObs.Alphabet().Exists(rtIt->second) )// FIXME: there seem to be relabeled transitions that do not exist    
               rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
        }
    } 
    // the controllable events that have been changed by the called function are set in the cGenerator cGenObs
    rGenObs.SetControllable(rControllableEvents);
}

// calcAbstAlphObsOCC(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTrans)
void calcAbstAlphObsOCC(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans)
{
    OP_DF("calcAbstAlphObsOCC(" << rGenObs.Name() << ", rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTRans)");
	// Initialization of variables
    rNewHighAlph = rHighAlph;
    rMapChangedTrans.clear();
    Generator genDyn(rGenObs);
    map<Transition,Transition> mapChangedTransReverse;
    vector<Idx> newPartitions;
    map<Idx,Idx> mapStateToPartition;
    map<Idx, EventSet> mapRelabeledEvents;
    bool done=false;
    #ifdef DF_PLOT
        Idx iterationCount=1;
        string name;
    #endif   
    // observer algorithm: In each step, a dynamic system is computed that fulfills the one-step observer condition
    // and output control consistency (OCC). 
	// Iterative application of the bisimulation algorithm and the relabeling procedure yields an automaton rGenObs
	// that, together with the new high-level alphabet rNewHighAlph fulfills the observer condition and OCC.
    while(done==false)
    {
        // compute dynamic system for Lm-observer and OCC on rGenObs
        calculateDynamicSystemObsOCC(rGenObs, rControllableEvents, rNewHighAlph, genDyn);
        #ifdef DF_PLOT
            name = ("./Automata/Plots/" + rGenObs.Name() + "DynamicSystem_" + ToString(iterationCount));
            genDyn.DotWrite(name);
        #endif

        Generator genPart;
        mapStateToPartition.clear();
        newPartitions.clear();
        // compute coarsest quasi-congruence on the dynamic system
        calcBisimulation(genDyn, mapStateToPartition, genPart, newPartitions);
        #ifdef DF_PLOT
            name = ("./Automata/Plots/" + rGenObs.Name() + "Bisimulation_" + ToString(iterationCount));
            genPart.DotWrite(name);
            ++iterationCount;
        #endif
    
        // check if quotient automaton is deterministic and free of unobservable events
		// and relabel transitions in rGenObs if necessary. The high-level alphabet is modified accordingly
        done=relabel(rGenObs, rControllableEvents, rNewHighAlph, newPartitions, mapStateToPartition, mapChangedTransReverse,  rMapChangedTrans, mapRelabeledEvents);
    }
}

// calculateDynamicSystemObsOCC(rGen, rControllableEvents, rHighAlph, rGenDyn)
void calculateDynamicSystemObsOCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn){
    OP_DF("calculateDynamicSystemObsOCC(" << rGen.Name() << "," << rControllableEvents.Name() << "," << rHighAlph.Name() << "," << rGenDyn.Name() << ")");
	// transition relation sorted in reverse order for backwards reachability
    TransSetX2EvX1 tset_X2EvX1;
    rGen.TransRel(tset_X2EvX1);

	// prepare generator rGenDyn
	rGenDyn.ClearTransRel();
	rGenDyn.InjectAlphabet(rHighAlph); // all high-level events are contained in the alphabet of rGenDyn
	// labels for transitions to marked states and for controllable paths
	std::string eventname = ( rGenDyn.EventSymbolTablep())->UniqueSymbol("cLabel_1");
	Idx cLabel = (rGenDyn.EventSymbolTablep())->InsEntry(eventname);
	eventname = ( rGenDyn.EventSymbolTablep())->UniqueSymbol("mLabel_1");
	Idx mLabel = (rGenDyn.EventSymbolTablep())->InsEntry(eventname);
	rGenDyn.InsEvent(cLabel);
	rGenDyn.InsEvent(mLabel);
	rGenDyn.InjectInitStates(rGen.InitStates() );
	rGenDyn.InjectStates(rGen.States() );
	rGenDyn.InjectMarkedStates(rGen.MarkedStates() );

	// maps for the construction of the dynamic system
	map<Idx, map<Idx, bool> > exitLocalStatesMap; // map from each exit state to locally backward reachable states and a boolean that is false if there exists an uncontrollable path to the exit state
	map<Idx, StateSet> entryLocalStatesMap; // map from entry states to locally forward reachable states
	StateSet::Iterator stIt, stEndIt;
	stIt = rGen.StatesBegin();
	stEndIt = rGen.StatesEnd();
	TransSet::Iterator tsIt, tsEndIt;
	bool isExitState;
	// go through all states of the original generator
	for(; stIt != stEndIt; stIt++){
	    OP_DF("calculateDynamicSystemObsOCC: loop over all states; current state: " << rGen.StateName(*stIt) 
		<< " [" << *stIt << "]");
		// determine the marked states that are locally reachable from the current state and insert 
		// transitions labeled with mLabel in rGenDyn
		forwardReachabilityObs(rGen, rHighAlph, *stIt, mLabel, rGenDyn);
		// if the current state is an exit state, carry out the backward reachability to determine 
		// which states can be reached on a controllable/uncontrollable path -> store in exitLocalStatesMap
		// in this case, also determine the corresponding entry states and compute their locally reachable states
		// for the entryLocalStatesMap
		tsIt = rGen.TransRelBegin(*stIt);
		tsEndIt = rGen.TransRelEnd(*stIt);
		isExitState = false;
		for( ; tsIt != tsEndIt; tsIt++){
			if(rHighAlph.Exists(tsIt->Ev) ){
               		 OP_DF("calculateDynamicSystemObsOCC: current state is an exit-state");
				isExitState = true;
				// if the local reach for the connected entry state has not been computed, yet, insert it in the 
				// entryLocalStatesMap
				if( entryLocalStatesMap.find(tsIt->X2) == entryLocalStatesMap.end() ){
					entryLocalStatesMap[tsIt->X2] = StateSet();
					LocalAccessibleReach(rGen,rHighAlph, tsIt->X2, entryLocalStatesMap[tsIt->X2]);
				}
			}
		}
		// if the current state is an exit state, compute the backward local reach with the controllability properties of the 
		// paths to locally backward reachable states
		if(isExitState == true){
			StateSet doneStates;
			exitLocalStatesMap[*stIt][*stIt] = false; // the exit state is reachable from the exit state via an uncontrollable path
			doneStates.Insert(*stIt);
			backwardReachabilityObsOCC(tset_X2EvX1, rControllableEvents, rHighAlph, *stIt, *stIt, false, exitLocalStatesMap, doneStates);
			
		}
		
	}
	// the generator rGenDyn is constructed by connecting all exit and entry states with their local state sets
	map<Idx, map<Idx, bool> >::const_iterator elIt, elEndIt;
	elIt = exitLocalStatesMap.begin();
	elEndIt = exitLocalStatesMap.end();
	map<Idx,bool>::const_iterator lcIt, lcEndIt;
	StateSet::Iterator exIt, exEndIt;
	map<Idx,StateSet>::const_iterator enIt;

	for( ; elIt != elEndIt; elIt++){
		lcEndIt = elIt->second.end();
		// go over all entry states reachable from the current exit state (via all feasible high-level events)
		tsIt = rGen.TransRel().Begin(elIt->first);
		tsEndIt = rGen.TransRel().End(elIt->first);
		for( ; tsIt != tsEndIt; tsIt++){
		    OP_DF("calculateDynamicSystemObsOCC: insert transitions for the high-level event" << rGen.EventName(tsIt->Ev) << "[" << tsIt->Ev << "]");
			if(rHighAlph.Exists(tsIt->Ev) ){
				bool controllable = rControllableEvents.Exists(tsIt->Ev);
				enIt = entryLocalStatesMap.find(tsIt->X2);
				stEndIt = enIt->second.End();
				// iterate over all locally backward reachable states from current exit state
				for(lcIt = elIt->second.begin(); lcIt != lcEndIt; lcIt ++){
					// iterate over all locally forward reachable states from current entry state
					for( stIt = enIt->second.Begin(); stIt != stEndIt; stIt++){
					    OP_DF("calculateDynamicSystemObsOCC: Transition added to resulting generator: " << 
        				rGenDyn.TStr(Transition(lcIt->first,tsIt->Ev,*stIt))); 
						
                        rGenDyn.SetTransition(lcIt->first,tsIt->Ev,*stIt); // insert a transition for each local state combination
                        if( controllable || lcIt->second ){ // insert an clabel transition if the local path is controllable or the high-level event is controllable
                            OP_DF("calculateDynamicSystemObsOCC: cLabel-Transition added to resulting generator: " << 
                            rGenDyn.TStr(Transition(lcIt->first,cLabel,*stIt))); 
                            
                            rGenDyn.SetTransition(lcIt->first,cLabel,*stIt);
                        }
					}
				}
			}
		}		
	}
}

// forwardReachabilityObs(rGen, rHighAlph, lowState, mLabel, rGenDyn)
void forwardReachabilityObs(const vGenerator& rGen, const EventSet& rHighAlph, Idx lowState, Idx mLabel, vGenerator& rGenDyn) {
	OP_DF("forwardReachabilityObs(" << rGen.Name() << "," << rHighAlph.Name() << "," << lowState << "," << rGenDyn.EventName(mLabel) << "," << rGenDyn.Name() << ")");
    // helpers:
	// iterators
	TransSet::Iterator tIt;
	TransSet::Iterator tEndIt;
	// todo list
	std::stack<Idx> todo;

	// algorithm: the locally reachable states from lowState are evaluated. If a reachable state is marked,
	// a transition with mLabel is inserted from lowState to that state is inserted in rGenDyn. 
	todo.push(lowState);
	StateSet doneStates;
	// if lowState is marked itself, the dynamic system contains a selfloop with the mlabel
	if(rGen.MarkedStates().Exists(lowState) ){
	    OP_DF("forwardReachabilityObs: Transition with mLabel added to resulting generator: " << 
        				rGenDyn.TStr(Transition(lowState,mLabel,lowState)));
		rGenDyn.SetTransition(lowState, mLabel, lowState);
	}

	doneStates.Insert(lowState);
	// the local reachability is evaluated until no new state is found
	while (! todo.empty()) {
		const Idx current = todo.top();
		todo.pop();
		tIt = rGen.TransRelBegin(current);
		tEndIt = rGen.TransRelEnd(current);
		for (; tIt != tEndIt; ++tIt) {
			// if the current transition is labeled with a high level event, it is skipped
			if (rHighAlph.Exists(tIt->Ev)) {
				continue;
			}
			// if the successor state has not been found, yst (not in doneStates)
			else if (! doneStates.Exists(tIt->X2)) {
				todo.push(tIt->X2);
				if(rGen.MarkedStates().Exists(tIt->X2) ){
				    OP_DF("forwardReachabilityObs: Transition with mLabel added to resulting generator: " << 
        				rGenDyn.TStr(Transition(lowState,mLabel,tIt->X2)));
					rGenDyn.SetTransition(lowState, mLabel, tIt->X2);
				}
				doneStates.Insert(tIt->X2);
			}
		}
	}
}

// backwardReachabilityObsOCC(rTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, currentState, controllablePath, rExitLocalStatesMap, rDoneStates) 
void backwardReachabilityObsOCC(const TransSetX2EvX1& rTransSetX2EvX1, const EventSet& rControllableEvents, const EventSet& rHighAlph, Idx exitState, Idx currentState, bool controllablePath, map<Idx, map<Idx, bool> >& rExitLocalStatesMap, StateSet& rDoneStates){
	OP_DF("backwardReachabilityObsOCC(rTransSetX2EvX1," << rControllableEvents.Name() << "," << rHighAlph.Name() << "," << exitState << "," << currentState << "," << controllablePath << ",rExitLocalStatesMap, rDoneStates)");
    // go along all backward transitions. Discard the goal state if it is reached via a high-level event or if it is in the rDoneStates and 
    // the controllability properties of the state do not change on the current path
    
    // helpers
	TransSetX2EvX1::Iterator tsIt, tsEndIt;
    tsIt = rTransSetX2EvX1.BeginByX2(currentState);
    tsEndIt = rTransSetX2EvX1.EndByX2(currentState);
	bool currentControllablePath;
	// we iterate over all backward transitions of the currentState to establish backward reachability
	for( ;tsIt != tsEndIt; tsIt++){
	    // states reachable via a high-level event are not in the local backward reach and the controllability property of the current exitState does not change
		if( !rHighAlph.Exists(tsIt->Ev) && tsIt->X1 != exitState){
			// if the state has not been visited, yet, the controllability of the current path are set in the rExitLocalStatesMap
            		if( !rDoneStates.Exists(tsIt->X1) ){
				rDoneStates.Insert(tsIt->X1);
				// the path is controllable if the current transition has a controllable event or the path was already controllable
				currentControllablePath = rControllableEvents.Exists(tsIt->Ev) || controllablePath;
				rExitLocalStatesMap[exitState][tsIt->X1] = currentControllablePath;
				// as the state has not been visited, yet, it is subject to a new backward reachability
				backwardReachabilityObsOCC(rTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, tsIt->X1, currentControllablePath, rExitLocalStatesMap, rDoneStates);
			}
			else{ // for an existing state, the controllability value can change from uncontrollable to controllable (if 
				// a new controllable path has been found). It is important to note, that the OCC condition implies that
				// if there is one controllable path, then the the state is flagged controllable except for the case of the 
				// given exitState that is always uncontrollable
				currentControllablePath = rControllableEvents.Exists(tsIt->Ev) || controllablePath;
				if(rExitLocalStatesMap[exitState][tsIt->X1] != currentControllablePath && currentControllablePath == true){
					rExitLocalStatesMap[exitState][tsIt->X1] = true;
					// as the controllabiity attribute of the current state changed it is subject to a new backward reachability
					backwardReachabilityObsOCC(rTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, tsIt->X1, true, rExitLocalStatesMap, rDoneStates);
				}
			}
		}
	}
}

// calcAbstAlphObsLCC(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents)
void calcAbstAlphObsLCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents){
    OP_DF("calcAbstAlphObsLCC(" << rGenObs.Name() << "," << "rHighAlph, rNewHighAlph, rMapRelabeledEvents)");
    // The controllable events are separated from the cGenerator. All functions that are successively 
    // called, are defined for vGenerators
    map<Transition,Idx>  changedtrans;
    EventSet rControllableEvents = rGenObs.ControllableEvents();
    // The function called next returns a relabeled generator and a map of relabeled transitions
    calcAbstAlphObsLCC(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, changedtrans);
    // for later use, the relabeled transitions are converted into relabeled events 
	// note that this function is accumulative, i.e., rMapRelabeledEvents need not be empty
    map<Transition,Idx>::iterator rtEndIt = changedtrans.end();
    map<Transition,Idx>::iterator rtIt =  changedtrans.begin();
    for(; rtIt != rtEndIt; rtIt++){
    if(rMapRelabeledEvents.find(rtIt->first.Ev) == rMapRelabeledEvents.end() ){
       rMapRelabeledEvents[rtIt->first.Ev] = set<Idx>();
       if(rGenObs.Alphabet().Exists(rtIt->second) )
           rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
    } else {  
       if(rGenObs.Alphabet().Exists(rtIt->second) )
           rMapRelabeledEvents[rtIt->first.Ev].insert(rtIt->second);
    }
      } 
      // the controllable events that have been changed by the called function are set in the cGenerator cGenObs
      rGenObs.SetControllable(rControllableEvents);
}

// calcAbstAlphObsLCC(rGenObs, rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTrans)
void calcAbstAlphObsLCC(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans)
{
    OP_DF("calcAbstAlphObsLCC(" << rGenObs.Name() << ", rControllableEvents, rHighAlph, rNewHighAlph, rMapChangedTRans)");
	// Initialization of variables
    rNewHighAlph = rHighAlph;
    rMapChangedTrans.clear();
    Generator genDyn(rGenObs);
    map<Transition,Transition> mapChangedTransReverse;
    vector<Idx> newPartitions;
    map<Idx,Idx> mapStateToPartition;
    map<Idx, EventSet> mapRelabeledEvents;
    bool done=false;
    #ifdef DF_PLOT
        Idx iterationCount=1;
        string name;
    #endif   
    // observer algorithm: In each step, a dynamic system is computed that fulfills the one-step observer condition
    // and local control consistency (LCC). 
	// Iterative application of the bisimulation algorithm and the relabeling procedure yields an automaton rGenObs
	// that, together with the new high-level alphabet rNewHighAlph fulfills the observer condition and LCC.
    while(done==false)
    {
         // compute dynamic system for Lm-observer and OCC on rGenObs
         calculateDynamicSystemObsLCC(rGenObs, rControllableEvents, rNewHighAlph, genDyn);
        #ifdef DF_PLOT
            name = ("./Automata/Plots/" + rGenObs.Name() + "DynamicSystem_" + ToString(iterationCount));
            genDyn.DotWrite(name);
        #endif

        Generator genPart;
        mapStateToPartition.clear();
        newPartitions.clear();
        // compute coarsest quasi-congruence on the dynamic system
        calcBisimulation(genDyn, mapStateToPartition, genPart, newPartitions);
        #ifdef DF_PLOT
            name = ("./Automata/Plots/" + rGenObs.Name() + "Bisimulation_" + ToString(iterationCount));
            genPart.DotWrite(name);
            ++iterationCount;
        #endif
    
        // check if quotient automaton is deterministic and free of unobservable events
		// and relabel transitions in rGenObs if necessary. The high-level alphabet is modified accordingly
        done=relabel(rGenObs, rControllableEvents, rNewHighAlph, newPartitions, mapStateToPartition, mapChangedTransReverse,  rMapChangedTrans, mapRelabeledEvents);
    }
}

// calculateDynamicSystemObsLCC(rGen, rControllableEvents, rHighAlph, rGenDyn)
void calculateDynamicSystemObsLCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn){
    OP_DF("calculateDynamicSystemObsLCC(" << rGen.Name() << "," << rControllableEvents.Name() << "," << rHighAlph.Name() << "," << rGenDyn.Name() << ")");
	// transition relation sorted in reverse order for backwards reachability
    TransSetX2EvX1 tset_X2EvX1;
   	rGen.TransRel(tset_X2EvX1);

   	// prepare generator rGenDyn
   	rGenDyn.ClearTransRel();
   	rGenDyn.InjectAlphabet(rHighAlph);
	std::string eventname = ( rGenDyn.EventSymbolTablep())->UniqueSymbol("ucLabel_1");
	Idx ucLabel = (rGenDyn.EventSymbolTablep())->InsEntry(eventname);
	eventname = ( rGenDyn.EventSymbolTablep())->UniqueSymbol("mLabel_1");
	Idx mLabel = (rGenDyn.EventSymbolTablep())->InsEntry(eventname);
	rGenDyn.InsEvent(ucLabel);
	rGenDyn.InsEvent(mLabel);
	rGenDyn.InjectInitStates(rGen.InitStates() );
	rGenDyn.InjectStates(rGen.States() );
	rGenDyn.InjectMarkedStates(rGen.MarkedStates() );

	// maps for the construction of the dynamic system
	map<Idx, map<Idx, bool> > exitLocalStatesMap; // map from each exit state to locally backward reachable states and a boolean that is false if there exists an uncontrollable path to the exit state
	map<Idx, StateSet> entryLocalStatesMap; // map from entry states to locally forward reachable states
	StateSet::Iterator stIt, stEndIt;
	stIt = rGen.StatesBegin();
	stEndIt = rGen.StatesEnd();
	TransSet::Iterator tsIt, tsEndIt;
	bool isExitState;
	// go through all states of the original generator
	for(; stIt != stEndIt; stIt++){
		OP_DF("calculateDynamicSystemObsLCC: loop over all states; current state: " << rGen.StateName(*stIt) 
		<< " [" << *stIt << "]");
		// determine the marked states that are locally reachable from the current state and insert 
		// transitions labeled with mLabel in rGenDyn
		forwardReachabilityObs(rGen, rHighAlph, *stIt, mLabel, rGenDyn);
		// if the current state is an exit state, carry out the backward reachability to determine 
		// which states can be reached on a controllable/uncontrollable path -> store in exitLocalStatesMap
		// in this case, also determine the corresponding entry states and compute their locally reachable states
		// for the entryLocalStatesMap
		tsIt = rGen.TransRelBegin(*stIt);
		tsEndIt = rGen.TransRelEnd(*stIt);
		isExitState = false;
		for( ; tsIt != tsEndIt; tsIt++){
			if(rHighAlph.Exists(tsIt->Ev) ){
			    OP_DF("calculateDynamicSystemObsLCC: current state is an exit-state");
				isExitState = true;
				// if the local reach for the connected entry state has not been computed, yet, insert it in the 
				// entryLocalStatesMap
				if( entryLocalStatesMap.find(tsIt->X2) == entryLocalStatesMap.end() ){
					entryLocalStatesMap[tsIt->X2] = StateSet();
					LocalAccessibleReach(rGen,rHighAlph, tsIt->X2, entryLocalStatesMap[tsIt->X2]);
				}
			}
		}
		// if the current state is an exit state, compute the backward local reach with the controllability properties of the 
		// paths to locally backward reachable states
		if(isExitState == true){
			StateSet doneStates;
			exitLocalStatesMap[*stIt][*stIt] = false; // the exit state is reachable from the exit state via an uncontrollable path
			doneStates.Insert(*stIt);
			backwardReachabilityObsLCC(tset_X2EvX1, rControllableEvents, rHighAlph, *stIt, *stIt, false, exitLocalStatesMap, doneStates);
			
		}
		
	}
	// the generator rGenDyn is constructed by connecting all exit and entry states with their local state sets
	map<Idx, map<Idx, bool> >::const_iterator elIt, elEndIt;
	elIt = exitLocalStatesMap.begin();
	elEndIt = exitLocalStatesMap.end();
	map<Idx,bool>::const_iterator lcIt, lcEndIt;
	StateSet::Iterator exIt, exEndIt;
	map<Idx,StateSet>::const_iterator enIt;

	for( ; elIt != elEndIt; elIt++){
		lcEndIt = elIt->second.end();
		// go over all entry states reachable from the current exit state (via all feasible high-level events)
		tsIt = rGen.TransRel().Begin(elIt->first);
		tsEndIt = rGen.TransRel().End(elIt->first);
		for( ; tsIt != tsEndIt; tsIt++){
		    OP_DF("calculateDynamicSystemObsLCC: insert transitions for the high-level event" << rGen.EventName(tsIt->Ev) << "[" << tsIt->Ev << "]");
			if(rHighAlph.Exists(tsIt->Ev) ){
				bool controllable = rControllableEvents.Exists(tsIt->Ev);
				enIt = entryLocalStatesMap.find(tsIt->X2);
				stEndIt = enIt->second.End();
				// iterate over all locally backward reachable states from current exit state
				for(lcIt = elIt->second.begin(); lcIt != lcEndIt; lcIt ++){
					// iterate over all locally forward reachable states from current entry state
					for( stIt = enIt->second.Begin(); stIt != stEndIt; stIt++){
						OP_DF("calculateDynamicSystemObsLCC: Transition added to resulting generator: " << 
        					rGenDyn.TStr(Transition(lcIt->first,tsIt->Ev,*stIt))); 
        				
                        			rGenDyn.SetTransition(lcIt->first,tsIt->Ev,*stIt); // insert a transition for each local state combination
						if( !(controllable || lcIt->second) ){ // insert an uclabel transition if the local path is uncontrollable
							OP_DF("calculateDynamicSystemObsLCC: cLabel-Transition added to resulting generator: " << 
                            				rGenDyn.TStr(Transition(lcIt->first,ucLabel,*stIt))); 

                            				rGenDyn.SetTransition(lcIt->first,ucLabel,*stIt);
						}
					}
				}
			}		
		}
	}
}

// backwardReachabilityObsLCC(crTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, currentState, controllablePath, rExitLocalStatesMap, rDoneStates)
void backwardReachabilityObsLCC(const TransSetX2EvX1& rTransSetX2EvX1, const EventSet& rControllableEvents, const EventSet& rHighAlph, Idx exitState, Idx currentState, bool controllablePath, map<Idx, map<Idx, bool> >& rExitLocalStatesMap, StateSet& rDoneStates){
	OP_DF("backwardReachabilityObsOCC(rTransSetX2EvX1," << rControllableEvents.Name() << "," << rHighAlph.Name() << "," << exitState << "," << currentState << "," << controllablePath << ",rExitLocalStatesMap, rDoneStates)");
    // go along all backward transitions. Discard the goal state if it is reached via a high-level event or if it is in the rDoneStates and 
    // the controllability properties of the state do not change on the current path
    
    // helpers
	TransSetX2EvX1::Iterator tsIt, tsEndIt;
    tsIt = rTransSetX2EvX1.BeginByX2(currentState);
    tsEndIt = rTransSetX2EvX1.EndByX2(currentState);
	bool currentControllablePath;
	// we iterate over all backward transitions of the currentState to establish backward reachability
	for( ;tsIt != tsEndIt; tsIt++){
	    // states reachable via a high-level event are not in the local backward reach and the controllability property of the current exitState does not change
		if( !rHighAlph.Exists(tsIt->Ev) && tsIt->X1 != exitState){
			// if the state has not been visited, yet, the controllability of the current path are set in the rExitLocalStatesMap
           if( !rDoneStates.Exists(tsIt->X1) ){
				rDoneStates.Insert(tsIt->X1);
				// the path is uncontrollable if the current transition has an uncontrollable event or the path was already uncontrollable
				currentControllablePath = rControllableEvents.Exists(tsIt->Ev) || controllablePath;
				rExitLocalStatesMap[exitState][tsIt->X1] = currentControllablePath;
				// as the state has not been visited, yet, it is subject to a new backward reachability
                backwardReachabilityObsLCC(rTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, tsIt->X1, currentControllablePath, rExitLocalStatesMap, rDoneStates);
			}
			else{ // for an existing state, the controllability value can change from controllable to uncontrollable (if 
				// a new uncontrollable path has been found). It is important to note, that the LCC condition implies that
				// if there is one uncontrollable path, then the state is flagged uncontrollable except for the case of the 
				// given exitState that is always uncontrollable
				currentControllablePath = rControllableEvents.Exists(tsIt->Ev) || controllablePath;
				if(rExitLocalStatesMap[exitState][tsIt->X1] != currentControllablePath && currentControllablePath == false){
					rExitLocalStatesMap[exitState][tsIt->X1] = currentControllablePath;
					// as the controllabiity attribute of the current state changed it is subject to a new backward reachability
                    backwardReachabilityObsLCC(rTransSetX2EvX1, rControllableEvents, rHighAlph, exitState, tsIt->X1, false, rExitLocalStatesMap, rDoneStates);
				}
			}
		}
	}
}

// relabel(rGenRelabel, rControllableEvents, rHighAlph, rNewPartitions, rMapStateToPartition, rMapChangedTransReverse, rMapChangedTrans, rMapRelabeledEvents)
bool relabel(vGenerator& rGenRelabel, EventSet& rControllableEvents, EventSet& rHighAlph, vector<Idx>& rNewPartitions, map<Idx,Idx>& rMapStateToPartition, map<Transition,Transition>& rMapChangedTransReverse, map<Transition,Idx>& rMapChangedTrans, map<Idx, EventSet>& rMapRelabeledEvents)
{

    OP_DF("relabel(" << rGenRelabel.Name() << "," << rControllableEvents.Name() << "," << rHighAlph.Name() << ", rNewPartition, rMapStateToPartition, rMapChangedTransReverse, rMapChangedTrans, rMapRelabeledEvents)");

	// helpers

	// keep track of how transitions between cosets have been relabeled
	// first Idx: coset from where transition starts, second Idx: high-level event,
	// third Idx: other coset, where transition ends, forth Idx: new high-level event
	map<Idx, map<Idx, map<Idx, Idx> > > mapRelabel;
	
	// set of low-level events, that have not been relabeled
	set<Idx> notRelabeledLowEvents;
	// transition relation of the original version of rGenRelabel
	const TransSet& transSet = rGenRelabel.TransRel();  
	TransSet::Iterator tIt = transSet.Begin();
	TransSet::Iterator tItEnd = transSet.End();
	// vector that contains pairs of original and changed transitions
	vector<pair<TransSet::Iterator,Transition> > editTransSet;
	// prepare data structure for keeping track of how transitions were relabeled
	vector<Idx>::iterator newPartIt = rNewPartitions.begin();
	vector<Idx>::iterator newPartItEnd = rNewPartitions.end();
	for(; newPartIt != newPartItEnd; ++newPartIt)
		mapRelabel[*newPartIt]=map<Idx, map <Idx, Idx> >();
	
	// algorithm: The relabeling according to the state partition as computed by the bisimulation algorithm is
	// performed. A high-level transitions is relebaled with a new event, if a transition with the same event leads
	// from the same coset to a different coset (non-determinism in the projected automaton). A low-level transition
	// is relabeled if it connects different cosets (corresponds to unobservable transition betweeen cosets). The 
	// also has build-in procedures to take care that as few new event labels as possible are introduced. 
	
	// iteration over all transitions of rGenRelabel
	for(; tIt != tItEnd; ++tIt)
	{
		OP_DF("relabel: current transition: " << rGenRelabel.TStr(*tIt) );
		Idx indexX1Partition=rMapStateToPartition[tIt->X1];
		OP_DF("relabel: X1 belongs to coset with index " << indexX1Partition);
		Idx indexX2Partition=rMapStateToPartition[tIt->X2];
		OP_DF("relabel: X2 belongs to coset with index " << indexX2Partition);
	
		//check if current transition is labeled with a high-level-event
		if(rHighAlph.Exists(tIt->Ev))
		{
			OP_DF("relabel: Event is high-level event");
			// In the first case, there exists no entry for this transition in mapRelabel. Hence, this is the
			// first occurence of tIt->Ev leaving the coset of tIt->X1 found in the iteration.
			// Thus, there is no need to relabel this event. Store transition in mapRelabel
			if(mapRelabel[indexX1Partition].find(tIt->Ev) == mapRelabel[indexX1Partition].end())
			{
				mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition]=tIt->Ev;
				OP_DF("relabel: First occurence of the current event leaving the current X1-coset");
			}
			// Otherwise, there exists an entry for the current event leaving the current X1-coset. It has to be
			// verified if the transition goes to the same coset as the previously found transitions or not.
			else 
			{
				// check if a transition with the same event has already been found that enters the same X2-coset. If
				// yes, the same label as for this transition is chosen. If not, a new label is introduced. 
				if(mapRelabel[indexX1Partition][tIt->Ev].find(indexX2Partition) != mapRelabel[indexX1Partition][tIt->Ev].end())
				{
					if(mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition] != tIt->Ev)
					{
						OP_DF("relabel: the same event leading to the same X2-coset has already been relabeld before. The current transition is relabeld correspondingly");
						pair<TransSet::Iterator,Transition> newPair;
						newPair.first=tIt;
						newPair.second = Transition(tIt->X1,mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition],tIt->X2);
						editTransSet.push_back(newPair);
					}
					else //nothing to be done
						OP_DF("relabel: There already exists a high-level transition from the current X1-coset to the same X2-coset and the current transition ist labeled with the same event. No relabeling necessary");
				}
				// there are entries of tIt->Ev, but no one leads to the same X2-coset. 
				// Check if new labels created before can be reused. If not, create a new one 
				// and eliminate the non-determinism by relebeling the current transition tIt with this event
				else 
				{
					bool createNewLabel = false;
					// check if the event already has been relabeled for some other transition
					if(rMapRelabeledEvents.find(tIt->Ev) != rMapRelabeledEvents.end())
					{
						EventSet::Iterator lsIt    = rMapRelabeledEvents[tIt->Ev].Begin();
						EventSet::Iterator lsItEnd = rMapRelabeledEvents[tIt->Ev].End();
						for(; lsIt != lsItEnd; ++lsIt)
						{
							createNewLabel = false;
							map<Idx,Idx>::iterator mapIt    = mapRelabel[indexX1Partition][tIt->Ev].begin();
							map<Idx,Idx>::iterator mapItEnd = mapRelabel[indexX1Partition][tIt->Ev].end();
							for(; mapIt != mapItEnd; ++mapIt)
							{	
								// if the currently investigated label has already been used in 
								// the current coset for a transition to another coset, relabeling is necessary
								if(mapIt->second == *lsIt)
								{
									createNewLabel = true;
									break;
								}
							}
							// use an old label if possible
							if(createNewLabel == false)
							{
								pair<TransSet::Iterator,Transition> newPair;
								newPair.first=tIt;
								newPair.second = Transition(tIt->X1,*lsIt,tIt->X2);
								editTransSet.push_back(newPair);
								OP_DF("relabel: An event was found that can be reused: " << rGenRelabel.EventName(*lsIt));
								break;
							}
						}
					}
					// if no useable labels are available, a new label has to be introduced
					else
					{
						createNewLabel = true;
					}
					// no label could be found that can be reused
					if(createNewLabel == true)
					{
						// create a new label and relabel the current transition with it
						std::string eventName = ( rGenRelabel.EventSymbolTablep())->UniqueSymbol(rGenRelabel.EventSymbolTablep()->Symbol(tIt->Ev) + "newHLevent_1");
						Idx newLabel = (rGenRelabel.EventSymbolTablep())->InsEntry(eventName);
				
						rGenRelabel.InsEvent(eventName);
						// if the original is controllable, the attribute of the new event is also controllable
						if(rControllableEvents.Exists(tIt->Ev)){ 
							rControllableEvents.Insert(newLabel);
						}
						OP_DF("relabel: No event that can be reused could be found. The new event " << rGenRelabel.EventName(newLabel) << " was created");
						rHighAlph.Insert(newLabel);
						//create the new transition and remember that the old one has to be replaced with the new one
						pair<TransSet::Iterator,Transition> newPair;
						newPair.first=tIt;
						newPair.second = Transition(tIt->X1,newLabel,tIt->X2);
						editTransSet.push_back(newPair);
						rMapRelabeledEvents[tIt->Ev].Insert(newLabel);
					}
				}
			}
		}
		// the current transition is labeled with a low-level event
		else
		{
			OP_DF("relabel: Event is low-level event");
			// potential relabeling is only necessary if the low-level transition leaves its coset
			if(indexX1Partition != indexX2Partition)
			{
				// transition connects two cosets and it is the first occurrence of the event. 
				// Create a new event and relabel the transition with it                                
				if(rMapRelabeledEvents.find(tIt->Ev) == rMapRelabeledEvents.end())
				{
					std::string eventName = ( rGenRelabel.EventSymbolTablep())->UniqueSymbol(rGenRelabel.EventSymbolTablep()->Symbol(tIt->Ev) + "newHLevent_1");
					Idx newLabel = (rGenRelabel.EventSymbolTablep())->InsEntry(eventName);
					rGenRelabel.InsEvent(eventName);
					// insert the new event in the high-level alphabet and set its controllability property
					rHighAlph.Insert(newLabel);
					if(rControllableEvents.Exists(tIt->Ev)){ 
					rControllableEvents.Insert(newLabel);
					}
					OP_DF("relabel: First occurence of current low-level event " << rGenRelabel.EventName(tIt->Ev) 
					<< " between cosets. The new event " << rGenRelabel.EventName(newLabel) << " was created");
					mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition]=newLabel;
					rMapRelabeledEvents[tIt->Ev].Insert(newLabel);
					pair<TransSet::Iterator,Transition> newPair;
					newPair.first=tIt;
					newPair.second  = Transition(tIt->X1,newLabel,tIt->X2);
					editTransSet.push_back(newPair);
				}
				// there are entries of tIt->Ev, but no one leads to the same X2-coset. 
				// Check if new labels created before can be reused. If not, create a new one 
				// and eliminate the non-determinism by relebeling the current transition tIt with this event
				else 
				{
					EventSet::Iterator lsIt    = rMapRelabeledEvents[tIt->Ev].Begin();
					EventSet::Iterator lsItEnd = rMapRelabeledEvents[tIt->Ev].End();
					bool createNewLabel = true;
					for(; lsIt != lsItEnd; ++lsIt)
					{
						bool labelFound = false;
						// check if the event already has been relabeled for some other transition
						if(mapRelabel.find(indexX1Partition) == mapRelabel.end())
						{
							// if the currently investigated label has already been used in 
							// the current coset for a transition to another coset, relabeling is necessary
							labelFound = true;
						}
						else
						{
							// label lsIt can be reused as this is the first low-level transition leaving the X1-coset
							if(mapRelabel[indexX1Partition].find(tIt->Ev) == mapRelabel[indexX1Partition].end())
							{
								labelFound = true;
							}
							else
							{
								// a transition with the same original event tIt->Ev leaving the X1-coset has been found before
								// and it entered a different X2-coset. Check if one of them was relabeled with lsIt. If yes, then lsIt can
								// not be reused and the next event has to be examined. If no, lsIt can be reused 
								if(mapRelabel[indexX1Partition][tIt->Ev].find(indexX2Partition) == mapRelabel[indexX1Partition][tIt->Ev].end())
								{
									map<Idx,Idx>::iterator mapIt=mapRelabel[indexX1Partition][tIt->Ev].begin();
									map<Idx,Idx>::iterator mapItEnd=mapRelabel[indexX1Partition][tIt->Ev].end();
									labelFound = true;
									for(; mapIt != mapItEnd; ++ mapIt)
									{
										if(mapIt->second== *lsIt)
										{
											labelFound = false;
											break;
										}
									}
								} 
								// a transition with the same original event tIt-Ev leaving the X1-coset has been found before
								// and it entered the same X2-coset. Reuse the corresponding label
								else
								{
									pair<TransSet::Iterator,Transition> newPair;
									newPair.first=tIt;
									newPair.second = Transition(tIt->X1,mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition],tIt->X2);
									editTransSet.push_back(newPair);
									OP_DF("relabel: A transition to same X2-coset has already been found and relabeled before; the current transition is also labeled with " << rGenRelabel.EventName(mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition]));
									createNewLabel = false;
									break;
								}
							}
						}
						// reuse existing event
						if(labelFound == true)
						{
							pair<TransSet::Iterator,Transition> newPair;
							newPair.first=tIt;
							newPair.second = Transition(tIt->X1,*lsIt,tIt->X2);
							editTransSet.push_back(newPair);
							OP_DF("relabele: Low level event "  << rGenRelabel.EventName(tIt->Ev) << " is relabeled with " << rGenRelabel.EventName(*lsIt) << " which can be reused");
							mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition]=*lsIt;
							createNewLabel = false;
							break;
						}
					} //end for
					// a new label has to be created and the corresponding transition is inserted
					if(createNewLabel == true)
					{
						std::string eventName = ( rGenRelabel.EventSymbolTablep())->UniqueSymbol(rGenRelabel.EventSymbolTablep()->Symbol(tIt->Ev) + "newHLevent_1"); //??
						Idx newLabel = (rGenRelabel.EventSymbolTablep())->InsEntry(eventName);
						rGenRelabel.InsEvent(eventName);
						// insert the new event in the high-level alphabet and set its controllability property
						rHighAlph.Insert(newLabel);
						if(rControllableEvents.Exists(tIt->Ev) ){ 
							rControllableEvents.Insert(newLabel);
						}
						OP_DF("relabel: No label could be reused, and the new event " << rGenRelabel.EventName(newLabel) << " was created");
						mapRelabel[indexX1Partition][tIt->Ev][indexX2Partition]=newLabel;
						pair<TransSet::Iterator,Transition> newPair;
						newPair.first=tIt;
						newPair.second = Transition(tIt->X1,newLabel,tIt->X2);
						editTransSet.push_back(newPair);
						rMapRelabeledEvents[tIt->Ev].Insert(newLabel);
					}
				}
			}
			// the low-level event does not have to be relabeled 
			else //indexX1Partition==indexX2Partition
			{
				notRelabeledLowEvents.insert(tIt->Ev);
			}
		}
	}
		
		
	// It is possible that events need not be relabeled. For example, if all occurrences of a low-level event have to 
	// be relabeled and added to the high-level alphabet, it is possible to just add the original low-level event
	// to the high-level alphabet without any relabeling. Note that this information is only available after all 
	// transitions have been evaluated.
	vector<pair<TransSet::Iterator,Transition> >::iterator etSetIt    = editTransSet.begin();
	vector<pair<TransSet::Iterator,Transition> >::iterator etSetItEnd = editTransSet.end();
	set<Idx> insertHighEvents;
	set<Idx> deleteHighEvents;
	bool quotAuotTrue=true;
	
	OP_DF("relabel: Trying to avoid unnecessariy new labels");
	// All relabeled transitions are checked
	for(etSetIt = editTransSet.begin(); etSetIt != editTransSet.end(); ++etSetIt)
	{
		Idx oldEvent = etSetIt->first->Ev;
		Idx newEvent = etSetIt->second.Ev;
		OP_DF("relabel: Checking transition X1="  << rGenRelabel.StateName(etSetIt->first->X1)<< " ["<<etSetIt->first->X1
		<< "] Ev=" << rGenRelabel.EventName(oldEvent) << " X2="  << rGenRelabel.StateName(etSetIt->first->X2)
		<< " [" << etSetIt->first->X2 << "] which shall be relabeled with event " << rGenRelabel.EventName(newEvent));
		// check if the original event is a low-level event and if there is an event left that has not been relabeled. If all events are relabeled, then at least one label can be replaced by the original event label.
		if(notRelabeledLowEvents.find(oldEvent) == notRelabeledLowEvents.end() && !rHighAlph.Exists(oldEvent))
		{
			OP_DF("relabel: There is no low-level event " << rGenRelabel.EventName(oldEvent) << " left");
			insertHighEvents.insert(oldEvent);
			// if a low-level event has been relabeled, the automaton under investigation is not the desired quotient automaton, yet.
			quotAuotTrue=false;
			// if newEvent is the first new event created for relabeling oldEvent, the relabeling is discarded
			if(rMapRelabeledEvents[oldEvent].Find(newEvent) == rMapRelabeledEvents[oldEvent].Begin())
			{
				OP_DF("relabel: Transition will not be relabeled");
				// if newEvent is the only event created for relabeling oldEvent, delete newEvent from rGenRelabel
				if(rMapRelabeledEvents[oldEvent].Find(newEvent)==(--(rMapRelabeledEvents[oldEvent].End())))
				{
					OP_DF("relabel: newEvent is the first and last event in rMapRelabeledEvents[oldEvent]");
					deleteHighEvents.insert(newEvent);
				}
				// delete pair<original Transition, new Transition>
				vector<pair<TransSet::Iterator,Transition> >::iterator helpIt = etSetIt;
				helpIt--;
				editTransSet.erase(etSetIt);
				etSetIt = helpIt;
			}
			else
			{
				// find predecessor of newEvent in map rMapRelabeledEvents[oldEvent] and use that event for relabeling
				Idx newLabel = *(--(rMapRelabeledEvents[oldEvent].Find(newEvent)));
				etSetIt->second.Ev = newLabel;
				// delete newEvent from rGenRelabel if it is no longer needed
				if(rMapRelabeledEvents[oldEvent].Find(newEvent)==(--(rMapRelabeledEvents[oldEvent].End())))
				{
					deleteHighEvents.insert(newEvent);
				}
			}
		}
	}
	// update the rMapRelabeledEvents with the changes
	// as the relabeling of oldEvent is discarded, it is removed from the map of relebeled events and
	// reinserted into the map of not relabeled events

	// mofidy alphabet of rGenRelabel and the abstraction alphabet by inserting the new high-level events and deleting the not needed new event labels.
	set<Idx>::iterator setIt    = insertHighEvents.begin();
	set<Idx>::iterator setItEnd = insertHighEvents.end();
	for(; setIt!= setItEnd; ++ setIt)
	{
		rHighAlph.Insert(*setIt);
	}
	setIt = deleteHighEvents.begin();
	setItEnd = deleteHighEvents.end();
	for(; setIt != setItEnd; ++setIt)
	{
		rGenRelabel.DelEvent(*setIt);
		rHighAlph.Erase(*setIt);
		rControllableEvents.Erase(*setIt); // controllable event is erased if it does not exist anymore. Schmidt 10/07
	}
		
	rControllableEvents = rControllableEvents * rGenRelabel.Alphabet(); // Schmidt 10/07
	// check if quotient automaton is deterministic and free of unobservable transitions, i.e., no transitions are relabeled and no low-level events are declared as high-level events
	if(editTransSet.empty()&& quotAuotTrue== true)
	{
		OP_DF("relabel: Leaving function relabel with true");
		return true; 
	}
		
	// delete original trnasitions and create relabeled transitions in rGenRelabel
	etSetIt = editTransSet.begin();
	etSetItEnd = editTransSet.end();
	for(; etSetIt != etSetItEnd; ++etSetIt)
	{
		map<Transition,Transition>::iterator mrIt;
		mrIt=rMapChangedTransReverse.find(*(etSetIt->first));
		// if the current relabeled transition has already been relabeled in a previous step of the observer algorithm, the original transition is found in the rMapChangedTransReverse map.
		if(mrIt!=rMapChangedTransReverse.end())
		{
			Transition originalTrans = mrIt->second;
			(rMapChangedTrans.find(originalTrans))->second = (etSetIt->second).Ev;
			rMapChangedTransReverse.erase(mrIt);
			rMapChangedTransReverse[etSetIt->second]=originalTrans;
			OP_DF("relabel: The transition X1= " << rGenRelabel.SStr((etSetIt->first)->X1) << " Ev=" << rGenRelabel.EStr((etSetIt->first)->Ev) << " X2= " << rGenRelabel.SStr((etSetIt->first)->X2) << " has already been relabeled in a former iteration step. The original transition was " << rGenRelabel.SStr(originalTrans.X1) << " Ev=" << rGenRelabel.EStr(originalTrans.Ev) << " X2= " << rGenRelabel.SStr(originalTrans.X2) << "; the new label is " << rGenRelabel.EStr((etSetIt->second).Ev));
		}
		// the current relabeled transition is simply inserted in the rMapChangedTrans map with its original transition if it is the first relabeling.
		else
		{
			rMapChangedTransReverse[etSetIt->second]=*(etSetIt->first);
			rMapChangedTrans[*(etSetIt->first)]=(etSetIt->second).Ev;
			OP_DF("relabel: First relabeling of transition X1=" << rGenRelabel.SStr((etSetIt->first)->X1) 
			<< " Ev=" << rGenRelabel.EStr((etSetIt->first)->Ev) << " X2= " << rGenRelabel.SStr((etSetIt->first)->X2) << " new label is " << rGenRelabel.EStr((etSetIt->second).Ev));
		}
		// The old transition is removed and the new transition is inserted into rGenRelabel
		rGenRelabel.ClrTransition(etSetIt->first);
		rGenRelabel.SetTransition(etSetIt->second);
	}
	
	OP_DF("relabel: leaving function with false");
	return false;
}

// insertRelabeledEvents(rGenPlant, rMapRelabeledEvents, rNewEvents) 
void insertRelabeledEvents(cGenerator& rGenPlant, const map<Idx,set<Idx> >&  rMapRelabeledEvents, cEventSet& rNewEvents) {
	map<Idx,set<Idx> >::const_iterator reEndIt = rMapRelabeledEvents.end();
	TransSet::Iterator tsEndIt = rGenPlant.TransRelEnd();
	TransSet::Iterator tsIt = rGenPlant.TransRelBegin();
	// check all transitions of rGenPlant
	for(; tsIt != tsEndIt; tsIt++){
		map<Idx,set<Idx> >::const_iterator reIt = rMapRelabeledEvents.find(tsIt->Ev);
		if(reIt == reEndIt) continue;
		const AttributeCFlags& attr = rGenPlant.EventAttribute(tsIt->Ev);
		set<Idx>::const_iterator rsEndIt = reIt->second.end();
		set<Idx>::const_iterator rsIt = reIt->second.begin();
		// if a transition with an event that gets new labels has been found, parallel transitions with the new 
		// events are added to rGenPlant
		for(; rsIt != rsEndIt; rsIt++){
			rGenPlant.InsEvent(*rsIt,attr);
			rNewEvents.Insert(*rsIt,attr);
			rGenPlant.SetTransition(tsIt->X1,*rsIt,tsIt->X2);
		}
	}
}

// insertRelabeledEvents(rGenPlant, rMapRelabeledEvents)
void insertRelabeledEvents(cGenerator& rGenPlant, const map<Idx,set<Idx> >&  rMapRelabeledEvents) {
	map<Idx,set<Idx> >::const_iterator reEndIt = rMapRelabeledEvents.end();
	TransSet::Iterator tsEndIt = rGenPlant.TransRelEnd();
	TransSet::Iterator tsIt = rGenPlant.TransRelBegin();
	// check all transitions of rGenPlant
	for(; tsIt != tsEndIt; tsIt++){
		map<Idx,set<Idx> >::const_iterator reIt = rMapRelabeledEvents.find(tsIt->Ev);
		if(reIt == reEndIt) continue;
		const AttributeCFlags& attr = rGenPlant.EventAttribute(tsIt->Ev);
		set<Idx>::const_iterator rsEndIt = reIt->second.end();
		set<Idx>::const_iterator rsIt = reIt->second.begin();
		// if a transition with an event that gets new labels has been found, parallel transitions with the new 
		// events are added to rGenPlant
		for(; rsIt != rsEndIt; rsIt++){
			rGenPlant.InsEvent(*rsIt,attr);
			rGenPlant.SetTransition(tsIt->X1,*rsIt,tsIt->X2);
		}
	}
}
 
}
