/** @file attributes.h Classes AttributeVoid and AttributeFlags  */


/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef ATTRIBUTES_H

#include "tokenreader.h"
#include "tokenwriter.h"

namespace faudes {


/**
 * Minimal Attribute. Attributes are used as template parameters for
 * faudes containers and generators and facilitate the modelling of customized  
 * properties of events, states and transitions.  The AttributeVoid class defines
 * the minimal interface of faudes attributes and therefore is the designated
 * base class for all attribute implementations. The AttributeVoid class does
 * not define any actual properties. See AttributeFlags for a non-trivial example.
 */

class AttributeVoid {
 public:

    /** Constructor */
    AttributeVoid(void) {};

    /** Destructor */
    virtual ~AttributeVoid(void) {};

    /**
     * Read attribute value from tokenreader.
     * This function skips all tokens until a String or Integer token indicates
     * the end of all attribute values. 
     * For derived classes, the read function shall shall try to read the attribute value and 
     * than call AttributeVoid::Read() to flush unknown tokens that presumably 
     * represent unsupported attributes. This function shall not throw token mismatch exceptions.
     *
     * @param rTr
     *   TokenReader to read from
     *
     * @exception Exception
     *   - IO error (id 1)
     */
    virtual void Read(TokenReader& rTr);
 
    /**
     * Write attribute value tokenreader
     *
     * @param rTw
     *   TokenWriter to write to
     *
     * @exception Exception
     *   - IO error (id 2)
     */
     virtual void Write(TokenWriter& rTw) const;

    /**
     * Write attribute value to console
     *
     */
     virtual void Write(void) const {};

    /**
     * Write attribute to string
     *
     * @return
     *  String representation of attribute
     * @exception Exception
     *   - IO error (id 2)
     *
     */
     virtual std::string ToString(void) const { return "";};

     /** 
      * Test for default value
      */
     bool  IsDefault(void) const {return true;};


}; // class AttributeVoid


/** Convenience typdef flag data */
typedef unsigned long int fType;

/**
 * Boolean flags Attribute. This attribute class uses a flag word to represent 
 * boolean values. The current implementation uses a long int type and hence handles
 * up to 32 flags. Each flag is accessed by the according bit mask. 
 *
 * The current version
 * of libFAUDES uses bits 0,1,2 of event attributes for controllability properties 
 * (see AttributeCFlags) and the bits 31 and 32 of state attributes for the Qt based
 * GUI project. Further versions may use the lower 8 bits of the event aflags and the higher 8
 * bits of state flags. You are free to use any flags for your application, but if possible try
 * to avoid the above mentioned. For extensive use of flags, you may also consider to
 * define a separate attribute class, perhaps with a different fType.
 */



class AttributeFlags : public AttributeVoid {
  public:

    /** Default constructor */
    AttributeFlags(void) : mFlags(mDefFlags) {};

    /** Destructor */
    virtual ~AttributeFlags(void) {};

    /**
     * Test a flag
     */
    bool Test(fType mask) const { return ( (mFlags & mask) != 0 ); };

    /**
     * Test multible flags, combine by "and"
     */
    bool TestAll(fType mask) const { return ( (mFlags & mask) == mask ); };

    /**
     * Test multible flags, combine by "or"
     */
    bool TestSome(fType mask) const { return ( (mFlags & mask) != 0 ); };

    /**
     * Test multible flags, combine by "not or"
     */
    bool TestNone(fType mask) const { return ( (mFlags & mask) == 0 ); };

    /**
     * Set multiple flags
     */
    void Set(fType mask) {mFlags |= mask; };

    /**
     * Clear multiple flags
     */
    void Clr(fType mask) {mFlags &= ~mask; };


   /**
    * Write attribute value to  tokenreader
    *
    * @param rTw
    *   TokenWriter to write to
    *
    * @exception Exception
    *   - IO error (id 2)
    */
    virtual void Write(TokenWriter& rTw) const;


    /**
     * Write attribute value to console
     *
     */
    virtual void Write(void) const;

   /**
    * Read attributes from tokenreader.
    * If the naxt token is a hex integer, read it and interpret as flag word. If not,
    * do nothing (dont flush tokens, dont throw errors)
    *
    * @param rTr
    *   tokenreader to read from
    *
    * @exception Exception
    *   - IO error (id 1)
    */
    virtual void Read(TokenReader& rTr);
 
    /**
     * Write attribute value to string
     *
     * @return
     *  string representation of attribute value
     *
     */
    virtual std::string ToString(void) const { return ToStringInteger16(mFlags);};

    /** 
     * Test for default value
     */
    bool  IsDefault(void) const {return mFlags==mDefFlags;};

    /** Flags (public access for convenience)  */
    fType mFlags;

    /* default flags */
    const static fType mDefFlags=0x0;


}; // class AttributeFlags


} // namespace faudes

#define ATTRIBUTES_H
#endif
