/** @file transset.h @brief Classes Transition, TTransSet and TaTransSet */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor

   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TRANSSET_H

#include "definitions.h"
#include "indexset.h"
#include "nameset.h"
#include "attributes.h"
#include <sstream>
#include <map>
#include <iostream>
#include "tokenreader.h"


namespace faudes {

/** @addtogroup ContainerClasses */
/** @{*/

/**
 * Triple (X1,Ev,X2) to represent current state, event and next state.
 * This class provides the triple in struct like fashion where the components
 * are of type faudes::Idx. While default order is lexographic, the transition
 * container TTransSet allows for alternative sorting. Since technically a
 * Transition is just a triple of indices, it is only the context of a generator
 * that actually makes it a transition (eg by providing mandatory symbolic event
 * names). 
 */
class Transition {

    public:

    /** Construct invalid Transition */
    Transition(void) : X1(0), Ev(0), X2(0) {};
    
    /**
     * Construct from values.
     * @param x1
     *   Index of current state
     * @param ev
     *   Index of Event
     * @param x2
     *   Index of next state
     */
    Transition(Idx x1, Idx ev,Idx x2) :
        X1(x1), Ev(ev), X2(x2) {}

    /** Default order for sorting container of Transition (lexographic) */
    inline bool operator < (const Transition& othertrans) const {
      if (X1 < othertrans.X1) return(true);
      if (X1 > othertrans.X1) return(false);
      if (Ev < othertrans.Ev) return(true);
      if (Ev > othertrans.Ev) return(false);
      if (X2 < othertrans.X2) return(true);
      return(false);
    } 

    /** Equality operator */
    inline bool operator == (const Transition& othertrans) const {
	  return ((X1 == othertrans.X1) && (Ev == othertrans.Ev) && (X2 == othertrans.X2));
    };
      
    /** Inequality operator */
    inline bool operator != (const Transition& othertrans) const {
	  return ((X1 != othertrans.X1) || (Ev != othertrans.Ev) || (X2 != othertrans.X2));
    };

    /** Test validity (all indices !=0 */
    inline bool Valid(void) const {
       return (X1!=0) && (Ev!=0) && (X2!=0);
    };
   
    /** Current state */
    Idx X1;
    
    /** Event */
    Idx Ev; 
    
    /** Next state */
    Idx X2;

    /** Pretty print to string */
    std::string Str(void) const {
      return ToStringInteger(X1)+"--("+ToStringInteger(Ev) + ")->" + ToStringInteger(X2);}
};




/**
 * Alternative ordering of Transitions. This class provides binary predicates
 * for sorting transition containers by variant lexographic order.
 */

class TransSort {

public:

/** Binary predicate for sorting transitions in order Ev, X1, X2 */
struct EvX1X2 {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.X2 < right.X2) return true;
    else return false;
  }
};

/** Binary predicate for sorting transitions in order Ev, X2, X1 */
struct EvX2X1 {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.X1 < right.X1) return true;
    else return false;
  }
};

/** Binary predicate for sorting transitions in order X2, Ev, X1 */
struct X2EvX1 {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X1 < right.X1) return true;
    else return false;
  }
};

/** Binary predicate for sorting transitions in order X2, X1, Ev */
struct X2X1Ev {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.Ev < right.Ev) return true;
    else return false;
  }
};

/** Binary predicate for sorting transitions in order X1, X2, Ev */
struct X1X2Ev {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.Ev < right.Ev) return true;
    else return false;
  }
};

/** Binary predicate for sorting transitions in order X1, Ev, X2 */
struct X1EvX2 {
  inline bool operator () (const Transition& left, const Transition& right) const {
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X2 < right.X2) return true;
    else return false;
  }
};

};


/** 
 * Set of Transitions.
 *
 * This container class provides similar functionality and interface as BaseSet, but
 * for Transitions rather than indices. The additional feature is a template parameter 
 * that allows to specify alternative sorting. For algorithms that examine a transition 
 * relation by e.g. the successor state X2, it may be worthwhile to copy a given 
 * TTransRel<TransSort::X1EvX2> to a TTransRel<TransSort::X2EvX1>. Example,
 * assuming some transition relation is given in default order 
 *
 * \code
 * TransSet transrel;                         // transrel default order X1-Ev-X2
 *
 * // ... some operation to set up transrel
 *
 * TTransSet<TransSort::X2EvX1> transByX2;    // alternative order X2-Ev-X1
 * transrel.ReSort(transByX2);                // copy and re-order transrel
 * Iterator tit    =transByX2.BeginByX2(x2)   // begin iterator, X2 specified 
 * Iterator tit_end=transByX2.EndByX2(x2)     // end iterator, X2 specified
 * for(; tit!=tit_end; ++tit) {               // loop over all transitions with specified x2
 *
 * // ... code to examine tramsitions with specified x2
 *
 * }
 *   
 * \endcode
 *
 * Note that it is the context of a Generator that
 * actually allows to interpret a TTransSet as a set of transitions as opposed to
 * a set of triples of indices. In particular, file IO of transitions is provided
 * by the generator class (although TTransSet provides output functions for debugging)
 */

template <class Cmp=TransSort::X1EvX2>
class TTransSet : public TBaseSet<Transition,Cmp> {
  public:

    /** @name Constructors & Destructor */ 
    /** @{ doxygen group */

    /** Construct an empty TTransSet object */
    TTransSet(void); 

    /**
     * Copy-constructor
     */
    TTransSet(const TTransSet& rOtherSet);


    /** Virtual destructor */
    virtual ~TTransSet() {};

    /** @} doxygen group */
    

    /** Iterator on transition */
    typedef typename TBaseSet<Transition, Cmp>::Iterator Iterator;


    /** @name Accessing individual transitions */ 
    /** @{ doxygen group */

    /** 
     * Add a Transition. 
     * 
     * @param rTransition 
     *   Reference to transition object
     * @return
     *   True if the transition was new to the set
     *
     */
    bool Insert(const Transition& rTransition);
    
    /** 
     * Add a Transition by indices
     *
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     * @param x2
     *   Successor state
     * @return
     *   True if the transition was new to the set
     *
     */
    bool Insert(Idx x1, Idx ev, Idx x2);
    
    /** 
     * Remove a Transition
     *
     * @return
     *   True if transition did exist
     */
    bool Erase(const Transition& t);
    
    /** 
     * Remove a Transition by x1, ev, x2
     *
     * @return
     *   True if transition did exist
     */
    bool Erase(Idx x1, Idx ev, Idx x2);

    /**
     * Remove a Transition by iterator
     *
     * @return
     *   Iterator to next transition
     */
    Iterator Erase(const Iterator& it);

    /**
     * Remove all transitions containing predecessor state x1
     *
     * @param x1
     *   State index
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    void EraseByX1(Idx x1);

    /**
     * Remove all transitions containing successor state x2.
     * This function iterates over all transitions to work with any sorting.
     * 
     * @param x2
     *   State index
     */
    void EraseByX2(Idx x2);

    /**
     * Remove all transitions containing event ev
     * This function iterates over all transitions to work with any sorting.
     *
     * @param ev
     *   Event index
     */
    void EraseByEv(Idx ev);
    
    /**
     * Remove all transitions containing state x
     * This function iterates over all transitions to work with any sorting.
     * 
     * @param x
     *   State index
     */
    void EraseByX1OrX2(Idx x);

    /**
     * Find transition given by x1, ev, x2
     *
     * 
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     * @param x2
     *   Successor state
     *
     * @return
     *   Iterator to transition or End() if not exists
     */
    Iterator Find(Idx x1, Idx ev, Idx x2) const;

    /**
     * Find specified transition
     *
     * 
     * @param t
     *   Transition
     *
     * @return
     *   Iterator to transition or End() if not exists
     */
    Iterator Find(const Transition& t) const;


    /**
     * Test existence
     *
     * @param t
     *   Transition
     *
     * @return
     *   bool
     */
    bool Exists(const Transition& t) const;

    /**
     * Test existence
     *
     * @param x1
     *   Predecessor state Idx
     * @param ev
     *   Event Idx
     * @param x2
     *   Successor state Idx
     *
     * @return
     *   bool
     */
    bool Exists(Idx x1, Idx ev, Idx x2) const;

    /**
     * Tests if a transition with specified predecessor or successor state
     * exists.
     *
     * @param x
     *   State Idx
     *
     * @return
     *   bool
     */
    bool Exists(Idx x) const;



    /** @} doxygen group */

    /** @name Transition iterators 
     *
     *  A variaty of iterators are provided to examine specified
     *  segments of the transition relation, e.g. all transitions starting
     *  from a given state. Note that implemetation of these iterators
     *  requires the transition set to be of sorted accordingly. Eg.
     *  scanning all transitions that are labled with a particular event
     *  requires a sorting TransSOrt::EvX1X2 orT ransSOrt::EvX2X1.
     *
     *
     */
 
    /** @{ doxygen group: iterators */

    /** 
     * Iterator to begin of set 
     *
     * @return
     *   TTransSet<Cmp>::Iterator
     */
    Iterator Begin(void) const;

    /** 
     * Iterator to end of set 
     *
     * @return
     *   TTransSet<Cmp>::Iterator
     */
    Iterator End(void) const;  

  
    /** 
     * Iterator to first Transition specified by current state.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - Sorting mismatch (id 68)
     */
    Iterator Begin(Idx x1) const;

    /** 
     * Iterator to end or Transitions with specified current state.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - Sorting mismatch (id 68)
     */
    Iterator End(Idx x1) const;

    /** 
     * Iterator to first Transitions specified by current state and event.
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::Iterator 
     *
     * @exception Exception
     *   - Sorting mismatch (id 68)
     */
    Iterator Begin(Idx x1, Idx ev) const;

    /** 
     * Iterator to first Transition after spcified current state and event. 
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator End(Idx x1, Idx ev) const;

    /**
     * Iterator to first Transition specified by event.
     * This function requires sorting TransSort::EvX1X2 or TransSort::EvX2X1. 
     * 
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator BeginByEv(Idx ev) const;

    /**
     * Iterator to first Transition after specified by event.
     * This function requires sorting TransSort::EvX1X2 or TransSort::EvX2X1 
     *
     * @param ev
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator EndByEv(Idx ev) const;

    /**
     * Iterator to first Transition specified by event and current state.
     * This function requires sorting TransSort::EvX1X2. 
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator BeginByEvX1(Idx ev, Idx x1) const;

    /**
     * Iterator to first Transition after specified ev and current state.
     * This function requires sorting TransSort::EvX1X2. 
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator EndByEvX1(Idx ev, Idx x1) const;

    /**
     * Iterator to first Transition specified by event and next state.
     * This function requires sorting TransSort::EvX2X1. 
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator BeginByEvX2(Idx ev, Idx x2) const;

    /**
     * Iterator to first Transition after specified event and next state.
     * This function requires sorting TransSort::EvX2X1. 
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator EndByEvX2(Idx ev, Idx x2) const;

    /**
     * Iterator to first Transition specified by successor state x2.
     * This function requires sorting TransSort::X2EvX1 or TransSort::X2X1Ev.  
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator BeginByX2(Idx x2) const;

    /**
     * Iterator to first Transition after specified successor state x2.
     * This function requires sorting TransSort::X2EvX1 or TransSort::X2X1Ev  
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator EndByX2(Idx x2) const;

    /**
     * Iterator to first Transition specified by successor x2  and ev.
     * This function requires sorting TransSort::X2EvX1.
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator BeginByX2Ev(Idx x2, Idx ev) const;

    /**
     * Iterator to first Transition after specified  successor x2  and ev.
     * This function requires sorting TransSort::X2EvX1.
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::Iterator
     *
     * @exception Exception
     *   - sorting mismatch (id 68)
     */
    Iterator EndByX2Ev(Idx x2, Idx ev) const;

    /** @} doxygen group */

    /** @name Misc */
    /** @{ doxygen group */

    /**
     * Get copy of trantision relation  sorted by other compare 
     * operator, e.g. TSort::X2EvX1
     * 
     * @return
     *   Transition relation
     */
    template<class OtherCmp>
    void ReSort(TTransSet<OtherCmp>& res) const;

    /** 
     * Get state space covered by transition set 
     *
     * @return
     *   Set of state indices used by some transition
     */
    StateSet StateSpace(void) const;

    /**
     * Get set of successor states for specified current state
     * 
     * @param x1
     *   Current state
     *
     * @return
     *   Set of state indices
     */
    StateSet StateSpaceX2(Idx x1) const;

    /**
     * Get set of successor states for specified current state and event
     * 
     * @param x1
     *   Current state
     * @param ev
     *   Event
     *
     * @return
     *   Set of state indices
     */
    StateSet StateSpaceX2(Idx x1, Idx ev) const;

    /**
     * Get set of events that are active for a specified current state
     * Since a transition set does not refer to a SymbolTable, this function
     * returns a set of plain indices. In order to interpret the set as an EventSet, 
     * the relevant SymbolTable must be supplied as second argument. If obmitting the second
     * argument, the defult SymbolTable is used.
     *
     * @param x1
     *   Current state
     * @param pSymTab
     *   SymbolTable to refer to
     *
     * @return 
     *   Set of events. 
     */
     EventSet ActiveEvents(Idx x1, SymbolTable* pSymTab=NULL) const;

    /** @} doxygen group */

protected:

  /** 
   * Write to TokenWriter, see TBaseSet for public wrappers.
   *
   * @param rTw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write, defaults to name of set
   *
   * @exception Exception
   *   - IO errors (id 2)
   */

   void DoWrite(TokenWriter& rTw, const std::string& rLabel="") const;


};

/** Type definition for default sorted TTransSet */
typedef TTransSet<TransSort::X1EvX2> TransSet;

/** Type definition for default sorted TTransSet */
typedef TTransSet<TransSort::X1EvX2> TransSetX1EvX2;

/** Type definition for ev, x1, x2 sorted TTransSet */
typedef TTransSet<TransSort::EvX1X2> TransSetEvX1X2;

/** Type definition for ev, x2, x1 sorted TTransSet */
typedef TTransSet<TransSort::EvX2X1> TransSetEvX2X1;

/** Type definition for x2, ev, x1 sorted TTransSet */
typedef TTransSet<TransSort::X2EvX1> TransSetX2EvX1;

/** Type definition for x2, x1, ev sorted TTransSet */
typedef TTransSet<TransSort::X2X1Ev> TransSetX2X1Ev;

/** Type definition for x1, x2, ev sorted TTransSet */
typedef TTransSet<TransSort::X1X2Ev> TransSetX1X2Ev;


/** 
 * Set of Transitions with attributes.
 *
 * This container class is derived from TTransSet to provide attributes as an
 * additional feature. As with TaBaseSet, the template parameter specifies the attribute class,
 * which in turn must provide some basic funtionality. In contrast to the TTransSet, the TaTransSet
 * is restricted to standard ordering. 
 *
 * Note that it is the context of a Generator that
 * actually allows to interpret a TaTransSet as a set of transitions as opposed to
 * a set of triples of indices with attributes. In particular, file IO of transitions is provided
 * by the generator class (although TaTransSet provides output functions for debugging)
 */


template <class Attr>
class TaTransSet : public TransSet {
  public:

   /** 
    * We implement "protected privacy for template classes" by friendship.
    * This is used for the pragmatic implemention conversion constructors.
    */
    template<class A> friend class TaTransSet;

    /** @name Constructors & Destructor */ 
    /** @{ doxygen group */

    /** Construct an empty TaTransSet object */
    TaTransSet(void); 

    /**
     * Copy-constructor (incl attributes) 
     */
    TaTransSet(const TaTransSet& rOtherSet);

    /**
     * Copy-Constructor (set attributes to default)
     */
    TaTransSet(const TTransSet<TransSort::X1EvX2>& rOtherSet);


    /** Virtual destructor */
    virtual ~TaTransSet() {}
    
    /** @} doxygen group */

    /** @name Basic maintenance */ 
    /** @{ doxygen group */

    /** Clear all transitions incl attributes */
    void Clear(void);

    /** 
     * clear all attributes 
     *
     */
    void ClearAttributes(void);

    /**
     * Get number of explicit (non-default) attributes
     *
     * @return 
     *   Number of entries in mAttributeMap
     */
    Idx AttributesSize(void) const;

    /** @} doxygen group */

    /** @name Accessing individual transitions */ 
    /** @{ doxygen group */
    
    /** 
     * Add a Transition by indices
     *
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     * @param x2
     *   Successor state
     *
     * @return
     *   True if the transition was new to the set
     */
    bool Insert(Idx x1, Idx ev, Idx x2);
    
    /** 
     * Add a Transition directly. If the transition already
     * exists, the attribute is maintained. Otherwise, the transition
     * is inserted with default attribute.
     * 
     * @param rTransition 
     *   Reference to transition object
     *
     * @return
     *   True if the transition was new to the set
     */
    bool Insert(const Transition& rTransition);
    
    /** 
     * Add a Transition with attribute.
     * 
     * @param rTransition 
     *   Reference to transition object
     * @param rAttr 
     *   Reference to attribute
     *
     * @return
     *   True if the transition was new to the set
     */
    bool Insert(const Transition& rTransition, const Attr& rAttr);
    
    /** 
     * Remove a Transition
     *
     * @return
     *   True if transition did exist
     */
    bool Erase(const Transition& t);
    
    /** 
     * Remove a Transition
     *
     * @return
     *   True if transition did exist
     */
    bool Erase(Idx x1, Idx ev, Idx x2);

    /**
     * Remove a Transition by iterator 
     *
     * @return
     *   Iterator to next transition
     */
    Iterator Erase(const Iterator& it);

    /**
     * Remove all transitions containing predecessor state x1.
     *
     * @param x1
     *   State index
     */
    void EraseByX1(Idx x1);

    /**
     * Remove all transitions containing successor state x2.
     *
     * @param x2
     *   State index
     */
    void EraseByX2(Idx x2);

    /**
     * Remove all transitions containing event ev.
     *
     * @param ev
     *   Event index
     */
    void EraseByEv(Idx ev);
    
    /**
     * Remove all transitions containing state x.
     * 
     * @param x
     *   State index
     */
    void EraseByX1OrX2(Idx x);

    /**
     * Get attribute reference by transition. 
     * Note that in order to produce a non-const 
     * reference this method will insert an explicit default attribute if necessary. 
     * If a const reference is sufficient, you should use Attribute(Idx) const instead.
     * 
     * @param rTrans
     *   Transition
     *
     * @return
     *   Transition attribute 
     */
     Attr* Attributep(const Transition& rTrans);

    /**
     * Get const reference to attribute by transition. 
     * 
     * @param rTrans
     *   Transition
     *
     * @return
     *   Transition attribute 
     */
     const Attr& Attribute(const Transition& rTrans) const;

    /** 
     * Set attribute. Sets the attribute of the sepcified element to the given value.  
     *
     * @param rTrans
     *   Transition to specify element
     *
     * @param attr
     *   Attribute value
     */
     void Attribute(const Transition& rTrans, const Attr& attr);

    /** 
     * Clear attribute to default
     *
     * @param rTrans
     *   Transition
     */
     void ClrAttribute(const Transition& rTrans);


    /** @} doxygen group */

    /** @name Misc */ 
    /** @{ doxygen group */

    /**
     * Copy to any other TaTransSet. 
     * This function ignores input attributes and sets the result attributes to the default value.
     *
     * @return
     *   TTransSet with no attributes
     */
    template<class OtherAttr> 
    void CopyWithoutAttributes(TaTransSet<OtherAttr>& res) const;

    /** @} doxygen group */

  protected:
    /** Stored attributes */
    std::map<Transition,Attr> mAttributeMap;

    /** default attribute */ 
    static Attr mDefAttribute; 

};

/** Statics per template parameter */
template<class Attr> Attr TaTransSet<Attr>::mDefAttribute;

/** @} doxygen group*/


/*
 *************************************************************************************************
 *************************************************************************************************
 * Implementation of transset without attributes
 *************************************************************************************************
 *************************************************************************************************
 */


/* convenience access to relevant scopes */
#define THIS TTransSet<Cmp>
#define TEMP template<class Cmp> 			
#define BASE TBaseSet<Transition,Cmp>



// TTransSet(void)
TEMP THIS::TTransSet(void) : BASE() {
  FD_DC("TTransSet(" << this << ")::TTransSet()");
}

// TTransSet(othertransrel)
TEMP THIS::TTransSet(const TTransSet& rOtherSet) : BASE(rOtherSet) {
  FD_DC("TTransSet(" << this << ")::TTransSet(rOtherSet "<< &rOtherSet <<")");
}

// iterator Begin() const
TEMP typename THIS::Iterator THIS::Begin(void) const {
  return BASE::Begin();
}

// iterator End() const
TEMP typename THIS::Iterator THIS::End(void) const {
  return BASE::End();
}


// Convenience macro for order typecheck
#define SORT_EXCEPTION  { std::stringstream errstr; \
  errstr << "transition set order mismatch " << std::endl; \
  throw Exception("TransSet::Iterator()", errstr.str(), 68); }


// iterator Begin(x1) const
TEMP typename THIS::Iterator THIS::Begin(Idx x1) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X1EvX2)) 
  if(typeid(Cmp)!=typeid(TransSort::X1X2Ev)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1,0,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator End(x1) const
TEMP typename THIS::Iterator THIS::End(Idx x1) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X1EvX2)) 
  if(typeid(Cmp)!=typeid(TransSort::X1X2Ev)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1+1,0,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator Begin(x1,ev) const
TEMP typename THIS::Iterator THIS::Begin(Idx x1, Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X1EvX2)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1,ev,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator End(x1,ev) const
TEMP typename THIS::Iterator THIS::End(Idx x1, Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X1EvX2)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1,ev+1, 0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator BeginByEv(ev) const
TEMP typename THIS::Iterator THIS::BeginByEv(Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX1X2)) 
  if(typeid(Cmp)!=typeid(TransSort::EvX2X1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator EndByEv(ev) const
TEMP typename THIS::Iterator THIS::EndByEv(Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX1X2)) 
  if(typeid(Cmp)!=typeid(TransSort::EvX2X1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev+1,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator BeginByEvX1(ev,x1) const
TEMP typename THIS::Iterator THIS::BeginByEvX1(Idx ev, Idx x1) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX1X2)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1,ev,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator EndByEvX1(ev,x1) const
TEMP typename THIS::Iterator THIS::EndByEvX1(Idx ev, Idx x1) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX1X2)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(x1+1,ev,0);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator BeginByEvX2(ev,x2) const
TEMP typename THIS::Iterator THIS::BeginByEvX2(Idx ev, Idx x2) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX2X1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev,x2);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator EndByEvX2(ev,x2) const
TEMP typename THIS::Iterator THIS::EndByEvX2(Idx ev, Idx x2) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::EvX2X1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev,x2+1);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator BeginByX2(x2) const
TEMP typename THIS::Iterator THIS::BeginByX2(Idx x2) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X2EvX1)) 
  if(typeid(Cmp)!=typeid(TransSort::X2X1Ev)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,0,x2);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator EndByX2(x2) const
TEMP typename THIS::Iterator THIS::EndByX2(Idx x2) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X2EvX1)) 
  if(typeid(Cmp)!=typeid(TransSort::X2X1Ev)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,0,x2+1);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator BeginByX2Ev(x2,ev) const
TEMP typename THIS::Iterator THIS::BeginByX2Ev(Idx x2, Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X2EvX1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev,x2);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}

// iterator EndByX2Ev(x2,ev) const
TEMP typename THIS::Iterator THIS::EndByX2Ev(Idx x2, Idx ev) const {
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X2EvX1)) 
  SORT_EXCEPTION;
#endif
  Transition tlx(0,ev+1,x2);
  return ThisIterator(BASE::pSet->lower_bound(tlx));
}



// DoWrite(rw,label)
TEMP void THIS::DoWrite(TokenWriter& rTw, const std::string& rLabel) const
{
  std::string label=rLabel;
  if(label=="") label=BASE::Name(); 
  rTw.WriteBegin(label);
  int oldcolumns = rTw.Columns();
  rTw.Columns(3);
  rTw.WriteBegin(label);

  Iterator tit;
  for (tit = Begin(); tit != End(); ++tit) {
    rTw << tit->X1; rTw << tit->Ev; rTw << tit->X2;
  }

  rTw.WriteEnd(label);
  rTw.Columns(oldcolumns);
}


// Insert(transition)
TEMP bool THIS::Insert(const Transition& t) {
  return BASE::Insert(t);
}

// Insert(x1,ev,x2)
TEMP bool  THIS::Insert(Idx x1, Idx ev, Idx x2) {
  FD_DC("TTransSet(" << this << ")::Insert(" << x1 << "-" << ev << "-" << x2 << ")");
  return BASE::Insert(Transition(x1, ev, x2));
}

// Erase(transition)
TEMP bool THIS::Erase(const Transition& t) {
  return BASE::Erase(t);  
}
    
// Erase(x1,ev,x2)
TEMP bool THIS::Erase(Idx x1, Idx ev, Idx x2) {
  FD_DC("TTransSet(" << this << ")::Erase(" << x1 << "-" << ev << "-" << x2 << ")");
  return BASE::Erase(Transition(x1, ev, x2));
}

// Erase(it)
TEMP typename THIS::Iterator THIS::Erase(const Iterator& it) {
  FD_DC("TTransSet(" << this << ")::Erase(" << it->X1 << "-" << it->Ev 
      << "-" << it->X2 << ")");
  return BASE::Erase(it);
}

// EraseByX1(x)
TEMP void THIS::EraseByX1(Idx x1) {
  FD_DC("TTransSet(" << this << ")::EraseByX1(" << x1 << ")");
#ifdef FAUDES_CHECKED
  if(typeid(Cmp)!=typeid(TransSort::X1EvX2)) 
  if(typeid(Cmp)!=typeid(TransSort::X1X2Ev)) 
  SORT_EXCEPTION;
#endif
  BASE::Detach();
  typename BASE::iterator lower, upper, it;
  Transition tl(x1,0,0);
  Transition tu(x1+1,0,0);
  lower = BASE::pSet->lower_bound(tl);
  upper = BASE::pSet->upper_bound(tu);
  BASE::pSet->erase(lower, upper);
}

// EraseByX2(x)
TEMP void THIS::EraseByX2(Idx x2) {
  FD_DC("TTransSet(" << this << ")::EraseByX2(" << x2 << ")");
  BASE::Detach();
  typename BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it != BASE::pSet->end();) {
    if (it->X2 == x2) {
      tmpit = it;
      it++;
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
}

// EraseByEv(ev)
TEMP void THIS::EraseByEv(Idx ev) {
  FD_DC("TTransSet(" << this << ")::EraseByEv(" << ev << ")");
  BASE::Detach();
  typename BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it != BASE::pSet->end();) {
    if (it->Ev == ev) {
      tmpit = it;
      it++;
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
}



// EraseByX1OrX2(x)
TEMP void THIS::EraseByX1OrX2(Idx x) {
  FD_DC("TTransSet(" << this << ")::EraseByX1OrX2(" << x << ")");
  BASE::Detach();
  typename BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it != BASE::pSet->end();) {
    if ((it->X1 == x) || (it->X2 == x)) {
      tmpit = it;
      it++;
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
  FD_DC("TTransSet(" << this << ")::EraseByX1OrX2(" << x << "): done");
}

// iterator Find(x1,ev,x2)
TEMP typename THIS::Iterator THIS::Find(Idx x1, Idx ev, Idx x2) const {
  return BASE::Find(Transition(x1,ev,x2));
}


// iterator Find(t)
TEMP typename THIS::Iterator THIS::Find(const Transition& t) const{
  return BASE::Find(t);
}

// Exists(t)
TEMP bool THIS::Exists(const Transition& t) const {
  return BASE::Exists(t);
}

// Exists(x1, ev, x2)
TEMP bool THIS::Exists(Idx x1, Idx ev, Idx x2) const {
  return BASE::Exists(Transition(x1,ev,x2));
}

// Exists(x)
TEMP bool THIS::Exists(Idx x) const {
  typename BASE::iterator it;
  for(it = BASE::pSet->begin(); it != BASE::pSet->end(); ++it) {
    if ((it->X1 == x) || (it->X2 == x)) {
      return true;
    }
  }
  return false;
}



// ReSort(res)
TEMP template<class OtherCmp>
void THIS::ReSort(TTransSet<OtherCmp>& res) const {
  Iterator it;
  res.Clear();
  for (it = Begin(); it != End(); ++it) {
    res.Insert(*it);
  }
}

// StateSpace()
TEMP StateSet THIS::StateSpace(void) const {
  StateSet states;
  Iterator it;
  for (it=Begin(); it!=End(); ++it) {
    states.Insert(it->X1);
    states.Insert(it->X2);
  }
  return states;
}

// StateSpaceX2(x1)
TEMP StateSet THIS::StateSpaceX2(Idx x1) const {
  StateSet states;
  Iterator it = Begin(x1);
  Iterator it_end = End(x1);
  while (it != it_end) {
    states.Insert(it->X2);
    ++it;
  }
  return states;
}

// StateSpaceX2(x1, ev)
TEMP StateSet THIS::StateSpaceX2(Idx x1, Idx ev) const {
  StateSet states;
  Iterator it = Begin(x1, ev);
  Iterator it_end = End(x1, ev);
  while (it != it_end) {
    states.Insert(it->X2);
    ++it;
  }
  return states;
}

// ActiveEvents(x1,pSymTab)
TEMP EventSet THIS::ActiveEvents(Idx x1, SymbolTable* pSymTab) const {
  Iterator it = Begin(x1);
  Iterator it_end = End(x1);
  EventSet result;
  if(pSymTab!=NULL) result.SymbolTablep(pSymTab);
  for (; it != it_end; ++it) {
    result.Insert(it->Ev);
  }
  return result;
}


#undef THIS
#undef TEMP
#undef BASE

/*
 *************************************************************************************************
 *************************************************************************************************
 * Implementation of transset with attributes
 *************************************************************************************************
 *************************************************************************************************
 */


/* convenience access to relevant scopes */
#define THIS TaTransSet<Attr>
#define TEMP template <class Attr>
#define BASE TTransSet<TransSort::X1EvX2>
        

// TaTransSet(void)
TEMP THIS::TaTransSet(void) : BASE() {
  FD_DC("TaTransSet(" << this << ")::TaTransSet()");
}

// TaTransSet(othertransrel)
TEMP THIS::TaTransSet(const TaTransSet& rOtherSet) : BASE(rOtherSet) {
  FD_DC("TaTransSet(" << this << ")::TaTransSet(rOtherSet "<< &rOtherSet <<")");
  mAttributeMap = rOtherSet.mAttributeMap;
}


// TaTransSet(othertransrel)
TEMP THIS::TaTransSet(const BASE& rOtherSet) :
  TransSet(rOtherSet)
{
  FD_DC("TaTransSet(" << this << ")::TaTransSet(rOtherSet "<< &rOtherSet <<")");
}

// CopyWithoutAttributes()
TEMP template<class OtherAttr> 
void THIS::CopyWithoutAttributes(TaTransSet<OtherAttr>& res) const{
  res.Clear();
  Assign(res);
}

// AttributesSize()
TEMP Idx THIS::AttributesSize(void) const {
  return (Idx)mAttributeMap.size();
}

// Clear()
TEMP void THIS::Clear(void) {
  FD_DC("TaTransSet(" << this << ")::Clear()");
  TransSet::Clear();
  mAttributeMap.clear();
}

// Insert(transition)
TEMP bool THIS::Insert(const Transition& t) {
  FD_DC("TaTransSet(" << this << ")::Insert(" << t.Str() << ")");
  // mAttributeMap.erase(t);
  return BASE::Insert(t);
}

// Insert(x1,ev,x2)
TEMP bool THIS::Insert(Idx x1, Idx ev, Idx x2) {
  FD_DC("TaTransSet(" << this << ")::Insert(" << x1 << "-" << ev << "-" << x2 << ")");
  Transition t(x1, ev, x2);
  // mAttributeMap.erase(t);
  return BASE::Insert(t);
}

// Insert(transition,attr)
TEMP bool THIS::Insert(const Transition& t, const Attr& attr) {
  if(attr.IsDefault())
    mAttributeMap.erase(t);
  else
    mAttributeMap[t]=attr;
  return BASE::Insert(t);
}

// Erase(transition)
TEMP bool THIS::Erase(const Transition& t) {
  bool ret= BASE::Erase(t);
  mAttributeMap.erase(t);
  return ret;
}
    
// Erase(x1,ev,x2)
TEMP bool THIS::Erase(Idx x1, Idx ev, Idx x2) {
  FD_DC("TaTransSet(" << this << ")::Erase(" << x1 << "-" << ev << "-" << x2 << ")");
  Transition t(x1, ev, x2);
  mAttributeMap.erase(t);
  return BASE::Erase(t);
}

// Erase(it)
TEMP typename THIS::Iterator THIS::Erase(const Iterator& it) {
#ifdef FAUDES_CHECKED
  if (it == End()) {
      std::stringstream errstr;
      errstr << "iterator out of range " << std::endl;
      throw Exception("TTransSet::Erase", errstr.str(), 69);
  }
#endif
  BASE::Detach();
  FD_DC("TaTransSet(" << this << ")::Erase(" << it->X1 << "-" << it->Ev 
      << "-" << it->X2 << ")");
  mAttributeMap.erase(*it);
  return BASE::Erase(it);
}

// EraseByX1(x)
TEMP void THIS::EraseByX1(Idx x1) {
  FD_DC("TaTransSet(" << this << ")::EraseByX1(" << x1 << ")");
  BASE::Detach();
  BASE::iterator lower, upper, it;
  Transition tl(x1,0,0);
  Transition tu(x1+1,0,0);
  lower = BASE::pSet->lower_bound(tl);
  upper = BASE::pSet->upper_bound(tu);
  if(mAttributeMap.size()!=0) 
    for(it=lower; it!=upper; ++it)
      mAttributeMap.erase(*it);
  BASE::pSet->erase(lower, upper);
}

// EraseByX2(x)
TEMP void THIS::EraseByX2(Idx x2) {
  FD_DC("TaTransSet(" << this << ")::EraseByX2(" << x2 << ")");
  BASE::Detach();
  BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it !=BASE::pSet->end();) {
    if (it->X2 == x2) {
      tmpit = it;
      it++;
      mAttributeMap.erase(*tmpit);
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
}

// EraseByEv(ev)
TEMP void THIS::EraseByEv(Idx ev) {
  FD_DC("TaTransSet(" << this << ")::EraseByEv(" << ev << ")");
  BASE::Detach();
  BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it !=BASE::pSet->end();) {
    if (it->Ev == ev) {
      tmpit = it;
      it++;
      mAttributeMap.erase(*tmpit);
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
}



// EraseByX1OrX2(x)
TEMP void THIS::EraseByX1OrX2(Idx x) {
  FD_DC("TaTransSet(" << this << ")::EraseByX1OrX2(" << x << ")");
  BASE::Detach();
  BASE::iterator it, tmpit;
  for(it = BASE::pSet->begin(); it !=BASE::pSet->end();) {
    if ((it->X1 == x) || (it->X2 == x)) {
      tmpit = it;
      it++;
      mAttributeMap.erase(*tmpit);
      BASE::pSet->erase(tmpit);
      continue;
    }
    it++;
  }
}

// Attributep(t)
TEMP Attr* THIS::Attributep(const Transition& rTrans) {
#ifdef FAUDES_CHECKED
  if (!Exists(rTrans)) {
    std::stringstream errstr;
    errstr << "transition " << rTrans.Str() << " not member of set" << std::endl;
    throw Exception("TTransSet::Attributep(t)", errstr.str(), 60);
  }
#endif
  typename std::map<Transition,Attr>::iterator ait;
  ait=mAttributeMap.find(rTrans);
  if( ait!=mAttributeMap.end() ) 
    return & (ait->second);
  // insert explicit default attribute
  typename std::map<Transition,Attr>::value_type entry(rTrans,mDefAttribute);
  ait=mAttributeMap.insert(entry).first;
  return & (ait->second);
}

// Attribute(t)
TEMP const Attr& THIS::Attribute(const Transition& rTrans) const {
#ifdef FAUDES_CHECKED
  if (!Exists(rTrans)) {
    std::stringstream errstr;
    errstr << "transition " << rTrans.Str() << " not member of set" << std::endl;
    throw Exception("TTransSet::Attribute(t)", errstr.str(), 60);
  }
#endif
  typename std::map<Transition,Attr>::const_iterator ait;
  ait=mAttributeMap.find(rTrans);
  if( ait!=mAttributeMap.end() ) {
    return ait->second;
  } 
  return mDefAttribute;
}

// Attribute(trans,attr)
TEMP void  THIS::Attribute(const Transition& rTrans, const Attr& rAttr) {
#ifdef FAUDES_CHECKED
  if (!Exists(rTrans)) {
    std::stringstream errstr;
    errstr << "transition " << rTrans.Str() << " not member of set" << std::endl;
    throw Exception("TTransSet::Attribute(t,attr)", errstr.str(), 60);
  }
#endif
  if(rAttr.IsDefault()) 
    mAttributeMap.erase(rTrans);
  else
    mAttributeMap[rTrans]=rAttr;
}

// ClearAttributes()
TEMP void  THIS::ClearAttributes(void) {
  mAttributeMap.clear();
}

#undef THIS
#undef TEMP
#undef BASE

#undef SORT_EXECPTION

} // namespace faudes




#define TRANSSET_H
#endif

